#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
When a user edits the attributes of a node using the Properties dialog and
presses `OK` the attributes correctness is checked by this module. Datatypes
and overflow conditions are checked. If no errors are detected the new set of
attributes is stored in the `PyTables` node. Otherwise the user is reported
about the error and the dialog remains open so the user can fix her mistake.
"""

__docformat__ = 'restructuredtext'

import tables
import numpy

from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils

translate = QtWidgets.QApplication.translate


def checkSyntax(value):
    """Check the syntax of a `Python` expression.

    :Parameters value: the Python expression to be evaluated
    """

    if value[0] in ("'", '"'):
        # Quotes are not permitted in the first position
        return False
    try:
        eval(value)
    except (ValueError, SyntaxError):
        return False
    else:
        return True


def formatStrValue(dtype, str_value):
    """
    Format the string representation of a value accordingly to its datatype.

    This function catches datatype errors for boolean values.

    :Parameters:

    - dtype: the value data type
    - str_value: the string representation of a value
    """

    if dtype == 'bool':
        # Every listed value is valid but none of them would
        # pass the out of range test later so we use a fake
        # valid value.
        # Beware that numpy.array(True).astype('bool')[()] fails
        # with a TypeError so we use '1' as a fake value
        if str_value in ['1', 'TRUE', 'True', 'true']:
            str_value = '1'
        elif str_value in ['0', 'FALSE', 'False', 'false']:
            str_value = '0'
        else:
            raise ValueError
    elif dtype.startswith('complex'):
        # Valid complex literal strings do not have parenthesis
        if str_value.startswith('(') and str_value.endswith(')'):
            str_value = str_value[1:-1]

    return str_value


def checkValue(dtype, str_value):
    """
    Check for overflow errors and dtype errors on integer and float values.

    By default, when overflow occurs in the creation of a ``numpy`` array,
    it is silently converted to the desired datatype:

    >>> numpy.array(-4).astype(numpy.uint8)
    array(252, dtype=uint8)
    >>> numpy.array(1420).astype(numpy.int8)
    array(-116, dtype=int8)

    This behavior can be acceptable for a library, but not for an end
    user application as ``ViTables`` so we have to catch such cases.

    :Parameters:

    - dtype: the value datatype
    - str_value: the string representation of a value
    """

    dtypes_map = {
        'int8': numpy.int8, 'int16': numpy.int16,
        'int32': numpy.int32, 'int64': numpy.int64,
        'uint8': numpy.uint8, 'uint16': numpy.uint16,
        'uint32': numpy.uint32, 'uint64': numpy.uint64,
        'float32': numpy.float32, 'float64': numpy.float64,
        }

    # For Python objects and strings no overflow can occur
    if dtype not in dtypes_map:
        return str_value

    new_array = numpy.array(str_value).astype(dtype)
    # Catches unexpected results from conversions
    # Examples: numpy.array('-23').astype('unint8') -> mismatch dtype
    # or numpy.array('99999').astype('int8') -> overflow
    if str(new_array[()]) != str_value:
        raise RuntimeWarning
    # If no errors are found return the original value
    return str_value


class AttrEditor(object):
    """
    Setup the attributes entered in the Properties dialog.

    When the user edits the Attributes Set Instance (see `PyTables` manual
    for details) of a given node via the Properties dialog and presses `OK`
    the validity of the new set of attributes is checked. If it is fine
    then the old `ASI` is replaced by the new one and the dialog is closed.
    If an error is found in the new set of attributes then the dialog
    remains opened until the user fixes the mistake.

    :Parameters:

    - `asi`: the Attributes Set Instance being updated
    - `title`: the TITLE attribute entered by the user in the Properties dialog
    - `user_table`: the table of user attributes edited by the user in the
        Properties dialog
    """

    def __init__(self, asi, title, user_table):
        """Initialise the attributes checker."""

        self.asi = asi

        # A dictionary with the attributes that have to be checked
        self.edited_attrs = {}
        model = user_table.model()
        rows = model.rowCount()
        # Parse the table and get string representations of its cell contents
        for row in range(0, rows):
            # As ViTables doesn't support editing ND-array attributes they
            # are marked in order to be found later
            name_item = model.item(row, 0)
            if not name_item.isEditable():
                multidim = True
            else:
                multidim = False
            name = model.item(row, 0).text()
            value = model.item(row, 1).text()
            dtype_index = model.indexFromItem(model.item(row, 2))
            dtype = user_table.indexWidget(dtype_index).currentText()
            self.edited_attrs[row] = (name, value, dtype, multidim)

        # Add the TITLE attribute to the dictionary
        if title is not None:
            self.edited_attrs[rows] = ('TITLE', title, 'string', False)


    def checkAttributes(self):
        """
        Check the user attributes table.

        If empty or repeated names, values mismatching the
        attribute type or out of range values are found then an error is
        reported. If the table is OK the node `ASI` is updated and the
        Properties dialog is closed.

        The attributes checked here are those of self.edited_attrs dict so the
        values are all string
        """

        # Error message for mismatching value/type pairs
        dtype_error = translate('AttrEditor', """\nError: "{0}" value """
            """mismatches its data type.""",
            'User attributes table editing error')

        # Error message for overflow values
        range_error = translate('AttrEditor', """\nError: "{0}" value """
            """is out of range.""",
            'User attributes table editing error')

        # Some times it's difficult to find out if the error
        # is due to a mismatch or to an overflow
        conversion_error = translate('AttrEditor', """\nError: "{0}" value """
            """is out of range or mismatches its data type.""",
            'User attributes table editing error')

        # Error message for syntax errors in Python attributes
        syntax_error = translate('AttrEditor', """\nError: "{0}" """
            """cannot be converted to a Python object.""",
            'User attributes table editing error')

        rows_range = self.edited_attrs.keys()

        # Check for empty Name cells
        for row in rows_range:
            name = self.edited_attrs[row][0]
            # Empty Value cells are acceptable for string attributes
            # but empty Name cells are invalid
            if name == '':
                return (False,
                        translate('AttrEditor',
                            "\nError: empty field Name in the row {0:d}",
                            'User attrs editing error').format(int(row + 1)))

        # Check for repeated names
        names_list = []
        for row in rows_range:
            name = self.edited_attrs[row][0]
            if not name in names_list:
                names_list.append(name)
            else:
                return (False,
                        translate('AttrEditor',
                            '\nError: attribute name "{0}" is repeated.',
                            'User attrs table editing error').format(name))

        # Check for dtype, range and syntax correctness of scalar attributes
        for row in rows_range:
            name, value, dtype, multidim = self.edited_attrs[row]
            if multidim == True:
                continue
            if dtype == 'python':
                # Check the syntax of the Python expression
                if not checkSyntax(value):
                    return (False, syntax_error.format(name))
            elif dtype == 'string':
                dtype = 'str'
            elif dtype == 'bytes':
                pass
            else:
                # Format properly the string representation of value
                # and check for dtype and overflow errors
                try:
                    value = formatStrValue(dtype, value)
                    value = checkValue(dtype, value)
                except OverflowError:
                    return (False, range_error.format(name))
                except ValueError:
                    return (False, dtype_error.format(name))
                except RuntimeWarning:
                    return (False, conversion_error.format(name))

            # If the attribute passes every test then its entry in the
            # dictionary of edited attributes is updated
            self.edited_attrs[row] = name, value, dtype, multidim

        return (True, None)


    def setAttributes(self):
        """
        Update edited attributes.

        This method is called when the user has successfully edited some
        attribute in the Properties dialog. If it happens then this method
        updates the `ASI` of the inspected `PyTables` node.
        """

        # Get rid of deleted attributes
        if 'TITLE' in self.edited_attrs:
            all_attrs = frozenset(self.asi._v_attrnamesuser + ["TITLE"])
        else:
            all_attrs = frozenset(self.asi._v_attrnamesuser)
        edited_attrs_names = frozenset([self.edited_attrs[row][0]
                                        for row in self.edited_attrs.keys()])
        for attr in (all_attrs - edited_attrs_names):
            try:
                self.asi._v_node._f_delattr(attr)
            except (tables.NodeError, AttributeError):
                vitables.utils.formatExceptionInfo()

        for row in self.edited_attrs.keys():
            # Scalar attributes are stored as
            # numpy scalar arrays of the proper type
            name, value, dtype, multidim = self.edited_attrs[row]
            if multidim == True:
                continue

            if dtype == 'python':
                value = eval('{0}'.format(value))
            elif dtype == 'bytes':
                # Remove the prefix and enclosing quotes
                value = value[2:-1]
                value = numpy.array(value).astype(dtype)[()]
            else:
                value = numpy.array(value).astype(dtype)[()]

            # Updates the ASI
            try:
                setattr(self.asi, name, value)
            except AttributeError:
                vitables.utils.formatExceptionInfo()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides editing capabilities for PyTables leaves and groups.

The module defines methods for deleting, cutting, copying, pasting and
moving nodes. Also group creation is supported.
"""

__docformat__ = 'restructuredtext'

import os

import tables

import vitables.utils



class TNodeEditor(object):
    """
    An editor for `tables.Leaf` and `tables.Group` instances.
    """

    def __init__(self, dbdoc):
        """
        Initialises the Leaf/Group editor.

        :Parameter h5file: the DBDoc instance being operated
        """

        if not dbdoc.hidden_group:
            dbdoc.createHiddenGroup()

        self.h5file = dbdoc.h5file
        self.hidden_group = dbdoc.hidden_group


    def delete(self, nodepath):
        """Delete a node.

        :Parameters:

        - `nodepath`: the full path of the node/link being deleted
        """

        try:
            self.h5file.remove_node(where=nodepath, recursive=True)
            self.h5file.flush()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()


    def cut(self, nodepath):
        """Moves a node to a hidden group of its database.

        The cut node must be stored somewhere or it will no be possible
        to paste it later. Storing it in the same database is extremely
        fast independently of the node size. Storing it in other database
        (i.e. in the temporary database) would have a cost which depends
        on the size of the cut node.

        :Parameters:

        - `nodepath`: the path of the node being cut
        """

        nodename = os.path.basename(nodepath)
        try:
            # The hidden group should contain at most 1 node
            for node in self.h5file.list_nodes(self.hidden_group):
                self.delete(node._v_pathname)

            self.move(nodepath, self, self.hidden_group, nodename)
            self.h5file.flush()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()


    def paste(self, src_node, parent, childname):
        """Copy a node to a different location.

        :Parameters:

        - `src_node`: the tables.Node instance being pasted
        - `parent`: the tables.Group where src_node will be pasted
        - `childname`: the new name of the node being pasted
        """

        try:
            self.h5file.copy_node(src_node, newparent=parent,
                newname=childname, overwrite=True, recursive=True)
            self.h5file.flush()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()


    def rename(self, nodepath, new_name):
        """
        Rename the selected node from the object tree.

        :Parameters:

        - `nodepath`: the full path of the node being renamed
        - `new_name`: the node new name
        """

        where, current_name = os.path.split(nodepath)

        try:
            self.h5file.rename_node(where, new_name, current_name, overwrite=1)
            self.h5file.flush()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()


    def create_group(self, where, final_name):
        """
        Create a new group under the given location.

        :Parameters:

        - `where`: the full path of the parent node
        - `final_name`: the new group name
        """

        try:
            if final_name in self.h5file.get_node(where)._v_children.keys():
                self.h5file.remove_node(where, final_name, recursive=True)
            self.h5file.create_group(where, final_name, title='')
            self.h5file.flush()
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()


    def move(self, childpath, dst_dbdoc, parentpath, childname):
        """Move a node to a different location.

        :Parameters:

        - `childpath`: the full path of the node being moved
        - `dst_dbdoc`: the destination database (a :meth:`DBDoc` instance)
        - `parentpath`: the full path of the new parent node
        - `childname`: the name of the node in its final location
        """

        try:
            dst_h5file = dst_dbdoc.h5file
            parent_node = dst_h5file.get_node(parentpath)
            if self.h5file is dst_h5file:
                self.h5file.move_node(childpath, newparent=parent_node,
                    newname=childname, overwrite=True)
            else:
                self.h5file.copy_node(childpath, newparent=parent_node,
                    newname=childname, overwrite=True, recursive=True)
                dst_h5file.flush()
                src_where, src_nodename = os.path.split(childpath)
                self.h5file.remove_node(src_where, src_nodename, recursive=1)
            self.h5file.flush()
            return childname
        except (tables.NodeError, OSError):
            vitables.utils.formatExceptionInfo()
            return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Site configuration module.

This module indicates the full install path of the ``ViTables`` package to the
running :meth:`vitables.vtapp.VTApp` instance.

Mac OS X boxes use the module as is.

Misc variables:

* __docformat__
* INSTALLDIR
* ICONDIR
* DOCDIR
* PLUGINSDIR
"""

import os.path

__docformat__ = 'restructuredtext'

INSTALLDIR = os.path.dirname(__file__)
ICONDIR = os.path.join(INSTALLDIR, "icons")
DOCDIR = os.path.join(INSTALLDIR, "htmldocs")
PLUGINSDIR = os.path.join(INSTALLDIR, "plugins")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Storing time series created with Pandas in PyTables. Example 1.
"""

import os

import numpy as np
import pandas as pd

# Create a DataFrame with a DateTimeIndex and linear data
dti = pd.date_range(start='1/1/2019', periods=365, name='Days')
ts = pd.Series(np.arange(1, 366), index=dti)
df = pd.DataFrame(ts)

# Create an empty HDFStore
output_dir = '../timeseries'
hdf5_name = 'pandas_test1.hdf5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
try:
    os.mkdir(output_dir)
except OSError:
    pass
finally:
    store = pd.HDFStore(filepath_hdf5)

# Store the dataframe as a PyTables Table under the root group
store.append('', df)
store.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a widget that wraps the view widget of leaves.

When a leaf node is opened in the tree of databases view the data stored in
that leaf will be displayed in the workspace using this wrapper widget.
"""

from qtpy import QtCore
from qtpy import QtWidgets

from . import leaf_view, leaf_model, df_model
from .. import utils as vtutils
from ..nodeprops import nodeinfo
from ..vtwidgets import zoom_cell

__docformat__ = 'restructuredtext'


class DataSheet(QtWidgets.QMdiSubWindow):
    """
    The widget containing the displayed data of a given dataset.

    :Parameter index: the index (in the tree of databases model) of the leaf
      whose data will be displayed
    """

    def __init__(self, index):
        """Display a given dataset in the MDI area.
        """

        # The main application window
        self.vtgui = vtutils.getVTApp().gui

        # The data structure (LeafNode/LinkNode instance) whose dataset
        # is being displayed
        dbt_model = self.vtgui.dbs_tree_model
        self.dbt_leaf = dbt_model.nodeFromIndex(index)

        # The tables.Node instance tied to that data structure
        pt_node = self.dbt_leaf.node
        if hasattr(pt_node, 'target'):
            # The selected item is a link and must be dereferenced
            leaf = pt_node()
        else:
            leaf = pt_node

        self.leaf_model = df_model.try_opening_as_dataframe(leaf)
        if not self.leaf_model:
            self.leaf_model = leaf_model.LeafModel(leaf)

        self.leaf_view = leaf_view.LeafView(self.leaf_model)

        super(DataSheet, self).__init__(self.vtgui.workspace,
                                        QtCore.Qt.SubWindow)
        self.setWidget(self.leaf_view)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        # Customize the title bar
        if not isinstance(leaf.title, str):
            title = str(leaf.title)
        else:
            title = leaf.title
        wtitle = "{0}\t{1}".format(self.dbt_leaf.name, title)
        self.setWindowTitle(wtitle)
        self.setWindowIcon(self.dbt_leaf.icon)

        # Eventually update the Node menu actions
        self.dbt_leaf.has_view = True
        self.vtgui.updateActions()

        self.pindex = QtCore.QPersistentModelIndex(index)

        # Connect signals to slots
        self.aboutToActivate.connect(self.syncTreeView)
        self.leaf_view.doubleClicked.connect(self.zoomCell)

    def closeEvent(self, event):
        """Close the window cleanly with the close button of the title bar.

        :Parameter event: the event being processed
        """

        # Ensure that Node menu actions are properly updated
        self.dbt_leaf.has_view = False
        self.vtgui.updateActions()

        # Propagate the event. In the process, self.widget().closeEvent
        # will be called
        QtWidgets.QMdiSubWindow.closeEvent(self, event)

        if not self.vtgui.workspace.subWindowList():
            self.vtgui.dbs_tree_view.setFocus(True)

    def focusInEvent(self, event):
        """Specialised handler for focus events.

        Synchronize with the tree view when the view gets keyboard focus.

        :Parameter event: the event being processed
        """

        # Sync the workspace with the tree view (if needed) but keep the
        # focus (giving focus to the tree view when a given view is
        # clicked is counter intuitive)
        QtWidgets.QMdiSubWindow.focusInEvent(self, event)
        self.syncTreeView()
        self.setFocus(True)

    def syncTreeView(self):
        """
        If the view is activated select its leaf in the tree of databases view.
        """

        if self.vtgui.editing_dlg is not None:
            self.vtgui.editing_dlg = None
            return
        # Locate the tree view leaf tied to this data sheet. Persistent
        # indices are used to get direct access to the leaf so we don't
        # have to walk the tree
        self.vtgui.dbs_tree_view.setCurrentIndex(
            QtCore.QModelIndex(self.pindex))

    def zoomCell(self, index):
        """Display the inner dimensions of a cell.

        :Parameter index: the index (in the leaf model) of the cell being zoomed
        """

        row = index.row()
        column = index.column()
        tmodel = index.model()
        data = tmodel.cell(row, column)
        if data is None:
            return

        # The title of the zoomed view
        node = self.dbt_leaf
        info = nodeinfo.NodeInfo(node)
        if node.node_kind == 'table':
            col = info.columns_names[column]
            title = '{0}: {1}[{2}]'.format(node.name, col,
                                           tmodel.start + row)
        else:
            title = '{0}: ({1},{2})'.format(node.name,
                                            tmodel.start + row, column)

        zoom_cell.ZoomCell(data, title, self.vtgui.workspace,
                           self.dbt_leaf)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a set of scripts by Francesc Alted.

"""A Table with nested records."""

import random

import tables

fileout = "nested_samples.h5"

# An example of enumerated structure
colors = tables.Enum(['red', 'green', 'blue'])

def write(h5file, desc, indexed):
    fileh = tables.open_file(h5file, "w")
    table = fileh.create_table(fileh.root, 'table', desc)
    #for colname in indexed:
    #    table.colinstances[colname].create_index()

    row = table.row
    for i in range(10):
        row['x'] = i
        row['y'] = 10.2-i
        row['z'] = i
        row['color'] = colors[random.choice(['red', 'green', 'blue'])]
        row['extra_info/name'] = "name%s" % i
        row['extra_info/info2/info3/z4'] =  i
        # All the rest will be filled with defaults
        row.append()

    fileh.close()

# The sample nested class description

class Info(tables.IsDescription):
    _v_pos = 2
    Name = tables.StringCol(16, dflt='sample string')
    Value = tables.Float64Col()

class Test(tables.IsDescription):
    """A description that has several columns"""
    x = tables.Int32Col(shape=2, dflt=0, pos=0)
    y = tables.Float64Col(dflt=1.2, shape=(2, 3))
    z = tables.UInt8Col(dflt=1)
    color = tables.EnumCol(colors, 'red', base='uint32', shape=(2,))
    Info = Info()
    class extra_info(tables.IsDescription):
        _v_pos = 1
        name = tables.StringCol(10)
        value = tables.Float64Col(pos=0)
        y2 = tables.Float64Col(dflt=1, shape=(2, 3), pos=1)
        z2 = tables.UInt8Col(dflt=1)
        class info2(tables.IsDescription):
            y3 = tables.Float64Col(dflt=1, shape=(2, 3))
            z3 = tables.UInt8Col(dflt=1)
            name = tables.StringCol(10)
            value = tables.EnumCol(colors, 'blue', base='uint32', shape=(1,))
            class info3(tables.IsDescription):
                name = tables.StringCol(10)
                value = tables.Time64Col()
                y4 = tables.Float64Col(dflt=1, shape=(2, 3))
                z4 = tables.UInt8Col(dflt=1)

# Write the file and read it
write(fileout, Test, ['info/info2/z3'])



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'calculator.ui'
#
# Created: Tue Jul  8 11:45:41 2014
#      by: PyQt4 UI code generator 4.10.3
#
# WARNING! All changes made in this file will be lost!

from qtpy import QtCore
from qtpy import QtWidgets

try:
    _encoding = QtWidgets.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtWidgets.QApplication.translate(context, text, disambig)

class Ui_CalculatorDialog(object):
    def setupUi(self, CalculatorDialog):
        CalculatorDialog.setObjectName("CalculatorDialog")
        CalculatorDialog.resize(640, 480)
        self.gridLayout = QtWidgets.QGridLayout(CalculatorDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.main_layout = QtWidgets.QHBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.save_layout = QtWidgets.QVBoxLayout()
        self.save_layout.setObjectName("save_layout")
        self.saved_label = QtWidgets.QLabel(CalculatorDialog)
        self.saved_label.setObjectName("saved_label")
        self.save_layout.addWidget(self.saved_label)
        self.saved_list =QtWidgets.QListWidget(CalculatorDialog)
        self.saved_list.setObjectName("saved_list")
        self.save_layout.addWidget(self.saved_list)
        self.list_buttons_layout = QtWidgets.QHBoxLayout()
        self.list_buttons_layout.setObjectName("list_buttons_layout")
        self.save_button = QtWidgets.QPushButton(CalculatorDialog)
        self.save_button.setObjectName("save_button")
        self.list_buttons_layout.addWidget(self.save_button)
        self.remove_button = QtWidgets.QPushButton(CalculatorDialog)
        self.remove_button.setObjectName("remove_button")
        self.list_buttons_layout.addWidget(self.remove_button)
        self.save_layout.addLayout(self.list_buttons_layout)
        self.main_layout.addLayout(self.save_layout)
        self.evaluation_layout = QtWidgets.QVBoxLayout()
        self.evaluation_layout.setObjectName("evaluation_layout")
        self.statements_label = QtWidgets.QLabel(CalculatorDialog)
        self.statements_label.setObjectName("statements_label")
        self.evaluation_layout.addWidget(self.statements_label)
        self.statements_edit = QtWidgets.QTextEdit(CalculatorDialog)
        self.statements_edit.setObjectName("statements_edit")
        self.evaluation_layout.addWidget(self.statements_edit)
        self.expression_label = QtWidgets.QLabel(CalculatorDialog)
        self.expression_label.setObjectName("expression_label")
        self.evaluation_layout.addWidget(self.expression_label)
        self.expression_edit = QtWidgets.QTextEdit(CalculatorDialog)
        self.expression_edit.setObjectName("expression_edit")
        self.evaluation_layout.addWidget(self.expression_edit)
        self.result_label = QtWidgets.QLabel(CalculatorDialog)
        self.result_label.setObjectName("result_label")
        self.evaluation_layout.addWidget(self.result_label)
        self.result_edit = QtWidgets.QLineEdit(CalculatorDialog)
        self.result_edit.setObjectName("result_edit")
        self.evaluation_layout.addWidget(self.result_edit)
        self.main_layout.addLayout(self.evaluation_layout)
        self.main_layout.setStretch(0, 1)
        self.main_layout.setStretch(1, 3)
        self.gridLayout.addLayout(self.main_layout, 0, 0, 1, 1)
        self.buttons = QtWidgets.QDialogButtonBox(CalculatorDialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Apply|QtWidgets.QDialogButtonBox.Close)
        self.buttons.setObjectName("buttons")
        self.gridLayout.addWidget(self.buttons, 1, 0, 1, 1)

        self.retranslateUi(CalculatorDialog)
        QtCore.QMetaObject.connectSlotsByName(CalculatorDialog)

    def retranslateUi(self, CalculatorDialog):
        CalculatorDialog.setWindowTitle(_translate("CalculatorDialog", "Evaluate expression", None))
        self.saved_label.setText(_translate("CalculatorDialog", "Saved expressions:", None))
        self.saved_list.setToolTip(_translate("CalculatorDialog", "<html><head/><body><p>Saved expressions.</p></body></html>", None))
        self.save_button.setToolTip(_translate("CalculatorDialog", "<html><head/><body><p>Store expression for future use.</p></body></html>", None))
        self.save_button.setText(_translate("CalculatorDialog", "Save", None))
        self.remove_button.setToolTip(_translate("CalculatorDialog", "<html><head/><body><p>Remove the selected expression from the saved expressions list.</p></body></html>", None))
        self.remove_button.setText(_translate("CalculatorDialog", "Remove", None))
        self.statements_label.setText(_translate("CalculatorDialog", "Statements:", None))
        self.statements_edit.setToolTip(_translate("CalculatorDialog", "<html><head/><body>\n"
"<p>Set of python statements that will be executed before the expression is evaluated. This can be used to import additional modules or do preliminary calculations to simplify the expression.</p>\n"
"</body></html>", None))
        self.expression_label.setText(_translate("CalculatorDialog", "Expression:", None))
        self.expression_edit.setToolTip(_translate("CalculatorDialog", "<html><head/><body>\n"
"<p>The results of the expression will be  saved in the result table. This field must contain a valid expression that returns a pytables array, list or scalar.</p>\n"
"<p>The expression can contain references to open data nodes. A reference can either be absolute and start from a file name or relative to the current group. Data refenrence is a string which is build by joining group names that lead to the data by dots. For example: &quot;filename.h5.some_group.sub_group.mydata&quot;. If &quot;sub_group&quot; is the current group then the string &quot;mydata&quot; can  be used as reference.</p>\n"
"</body></html>", None))
        self.result_label.setText(_translate("CalculatorDialog", "Result table:", None))
        self.result_edit.setToolTip(_translate("CalculatorDialog", "<html><head/><body>\n"
"<p>Reference to the destination for the expression result. Data refenrence is a string which is build by joining group names that lead to the data by dots. For example: &quot;filename.h5.some_group.sub_group.myresult&quot;. If &quot;sub_group&quot; is the current group then the string &quot;myresult&quot; can  be used as reference.</p>\n"
"<p>The result table must not exitst.</p>\n"
"</body></html>", None))




"""Test class for vtgui.py"""

import pytest

from qtpy import QtCore
from qtpy import QtWidgets

import vitables.logger as logger


@pytest.mark.usefixtures('launcher')
class TestVTGui(object):
    def test_dockWidget(self, launcher):
        logger_dock = launcher.gui.findChild(QtWidgets.QDockWidget,
                                             'LoggerDockWidget')
        # Does the dock widget exist?
        assert logger_dock
        # Is it docked to the proper area?
        assert (launcher.gui.dockWidgetArea(logger_dock) ==
                QtCore.Qt.BottomDockWidgetArea)
        # Does it have the required features?
        assert (logger_dock.features() ==
                QtWidgets.QDockWidget.DockWidgetClosable
                | QtWidgets.QDockWidget.DockWidgetMovable
                | QtWidgets.QDockWidget.DockWidgetFloatable)
        # Is the logger widget properly set?
        assert isinstance(logger_dock.widget(), logger.Logger)

    def test_hsplitter(self, launcher):
        hsplitter = launcher.gui.centralWidget().findChild(QtWidgets.QSplitter,
                                                           'hsplitter')
        assert hsplitter
        assert hsplitter.count() == 2
        assert hsplitter.indexOf(launcher.gui.dbs_tree_view) == 0
        assert hsplitter.indexOf(launcher.gui.workspace) == 1

    def test_actions(self, launcher):
        gui_actions = launcher.gui.gui_actions.keys()
        expected_actions = \
            ['fileNew', 'fileOpen', 'fileOpenRO', 'fileClose', 'fileCloseAll',
             'fileSaveAs', 'fileExit', 'nodeOpen', 'nodeClose',
             'nodeProperties', 'nodeNew', 'nodeRename', 'nodeCut', 'nodeCopy',
             'nodePaste', 'nodeDelete', 'queryNew', 'queryDeleteAll',
             'settingsPreferences', 'windowCascade', 'windowTile',
             'windowRestoreAll', 'windowMinimizeAll', 'windowClose',
             'windowCloseAll', 'windowSeparator', 'mdiTabbed',
             'helpUsersGuide', 'helpAbout', 'helpAboutQt', 'helpVersions',
             'calculate']
        assert sorted(gui_actions) == sorted(expected_actions)

    def test_fileToolBar(self, launcher):
        file_tb = launcher.gui.findChild(QtWidgets.QToolBar, 'File toolbar')
        assert file_tb

        tb_actions = [a.objectName() for a in file_tb.actions()]
        expected_actions = ['fileNew', 'fileOpen', 'fileClose', 'fileSaveAs']
        assert sorted(tb_actions) == sorted(expected_actions)

    def test_nodeToolBar(self, launcher):
        node_tb = launcher.gui.findChild(QtWidgets.QToolBar, 'Node toolbar')
        assert node_tb

        tb_actions = [a.objectName() for a in node_tb.actions()]
        expected_actions = ['nodeNew', 'nodeCut', 'nodeCopy', 'nodePaste',
                            'nodeDelete']
        assert sorted(tb_actions) == sorted(expected_actions)

    def test_queryToolBar(self, launcher):
        query_tb = launcher.gui.findChild(QtWidgets.QToolBar, 'Query toolbar')
        assert query_tb

        tb_actions = [a.objectName() for a in query_tb.actions()]
        expected_actions = ['queryNew', 'queryDeleteAll']
        assert sorted(tb_actions) == sorted(expected_actions)

    def test_helpToolBar(self, launcher):
        help_tb = launcher.gui.findChild(QtWidgets.QToolBar, 'Help toolbar')
        assert help_tb

        tb_actions = [a.objectName() for a in help_tb.actions()]
        expected_actions = ['helpUsersGuide', 'whatis_help_toolbar']
        assert sorted(tb_actions) == sorted(expected_actions)

    def test_statusBarWidget(self, launcher):
        sbw = launcher.gui.statusBar().findChild(QtWidgets.QLabel,
                                                 'status bar widget')
        assert sbw
        sbw_sp = sbw.sizePolicy()
        hsp, vsp = sbw_sp.horizontalPolicy(), sbw_sp.verticalPolicy()
        assert hsp == QtWidgets.QSizePolicy.MinimumExpanding
        assert vsp == QtWidgets.QSizePolicy.Minimum

    @pytest.fixture()
    def menuBar(self, launcher):
        return launcher.gui.menuBar()

    def test_menus(self, menuBar):
        assert len(menuBar.actions()) == 6

    def test_fileMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'file_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['fileNew', 'fileOpen', 'fileOpenRO', 'fileClose',
                            'fileCloseAll', 'fileSaveAs', 'fileExit']
        assert sorted(actions) == sorted(expected_actions)

        menus = [a.menu().objectName() for a in menu_actions if a.menu()]
        assert sorted(menus) == ['import_csv_submenu', 'open_recent_submenu']

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 4

    def test_nodeMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'node_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['nodeOpen', 'nodeClose', 'nodeProperties',
                            'nodeNew', 'nodeRename', 'nodeCut', 'nodeCopy',
                            'nodePaste', 'nodeDelete']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 1

    def test_datasetMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'dataset_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['queryNew', 'calculate', 'export_csv']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 1

    def test_settingsMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'settings_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['settingsPreferences']
        assert sorted(actions) == sorted(expected_actions)

        menus = [a.menu().objectName() for a in menu_actions if a.menu()]
        assert sorted(menus) == ['settings_toolbars_submenu']

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 1

    def test_windowMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'window_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['windowCascade', 'windowTile',
                           'windowRestoreAll', 'windowMinimizeAll',
                           'windowClose', 'windowCloseAll', 'mdiTabbed']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 2

    def test_helpMenu(self, menuBar):
        menu = menuBar.findChild(QtWidgets.QMenu, 'help_menu')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['helpUsersGuide', 'helpAbout', 'helpAboutQt',
                        'helpVersions', 'whatis_help_menu']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 2

    def test_viewCM(self, launcher):
        menu = launcher.gui.findChild(QtWidgets.QMenu, 'view_cm')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['fileNew', 'fileOpen', 'fileOpenRO', 'fileClose',
                            'fileCloseAll', 'fileSaveAs', 'fileExit']
        assert sorted(actions) == sorted(expected_actions)

        menus = [a.menu().objectName() for a in menu_actions if a.menu()]
        assert sorted(menus) == ['import_csv_submenu', 'open_recent_submenu']

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 4

    def test_rootNodeCM(self, launcher):
        menu = launcher.gui.findChild(QtWidgets.QMenu, 'root_node_cm')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['fileClose', 'fileSaveAs', 'nodeProperties',
                            'nodeNew', 'nodeCopy', 'nodePaste',
                            'queryDeleteAll']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 3

    def test_groupNodeCM(self, launcher):
        menu = launcher.gui.findChild(QtWidgets.QMenu, 'group_node_cm')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['nodeProperties', 'nodeNew', 'nodeRename',
                            'nodeCut', 'nodeCopy', 'nodePaste', 'nodeDelete']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 1

    def test_leafNodeCM(self, launcher):
        menu = launcher.gui.findChild(QtWidgets.QMenu, 'leaf_node_cm')
        menu_actions = menu.actions()
        assert menu

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['nodeOpen', 'nodeClose',  'nodeProperties',
                            'nodeRename', 'nodeCut', 'nodeCopy', 'nodePaste',
                            'nodeDelete', 'queryNew', 'export_csv']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 4

    def test_mdiCM(self, launcher):
        menu = launcher.gui.findChild(QtWidgets.QMenu, 'mdi_cm')
        menu_actions = menu.actions()
        assert menu

        menus = [a.menu().objectName() for a in menu_actions if a.menu()]
        assert sorted(menus) == ['window_menu']



"""Function that calls eval for calculator.

It is separated into a module in order to import * from numpy and thus
simplify writing expressions. (Such import is not allowed inside a
function.)

"""

from numpy import *


def evaluate(statements, expression, globals_dict):
    """Evaluate expression and return results."""
    exec(statements)
    globals_dict.update(globals())
    globals_dict.update(locals())
    return eval(expression, globals_dict)



"""Test class for utils.py"""

import pytest

from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils as utils


@pytest.mark.usefixtures('launcher')
class TestUtils(object):
    def test_getVTApp(self):
        vtapp = utils.getVTApp()
        assert vtapp.objectName() == 'VTApp'

    def test_getGui(self):
        gui = utils.getGui()
        assert gui.objectName() == 'VTGUI'

    def test_getModel(self):
        model = utils.getModel()
        assert model.objectName() == 'dbs_tree_model'

    def test_getView(self):
        view = utils.getView()
        assert view.objectName() == 'dbs_tree_view'

    def test_getSelectedIndexes(self):
        pass

    def test_getSelectedNodes(self):
        pass

    @pytest.fixture()
    def actions(self):
        # Menu to be enlarged
        menubar = utils.getGui().menuBar()
        help_menu = menubar.findChild(QtWidgets.QMenu, 'help_menu')

        # Actions to insert/append
        new_action = QtWidgets.QAction('TestAction')
        new_action.setObjectName('testaction')
        new_menu = QtWidgets.QMenu('TestMenu')
        new_menu.setObjectName('testmenu')

        return {
            'help_menu': help_menu,
            'new_action': new_action,
            'new_menu': new_menu,
        }

    def test_insertInMenu(self, actions):
        uid = 'helpUsersGuide'

        # Insert a new action atop of the menu
        utils.insertInMenu(actions['help_menu'], actions['new_action'], uid)
        hm_actions = actions['help_menu'].actions()
        assert (hm_actions[0].objectName() ==
                actions['new_action'].objectName())
        actions['help_menu'].removeAction(hm_actions[0])

        # Insert a new menu atop of the menu
        utils.insertInMenu(actions['help_menu'], actions['new_menu'], uid)
        hm_actions = actions['help_menu'].actions()
        assert (hm_actions[0].menu().objectName() ==
                actions['new_menu'].objectName())
        actions['help_menu'].removeAction(hm_actions[0])

    def test_addToMenu(self, actions):
        # Append a new action
        utils.addToMenu(actions['help_menu'], actions['new_action'])
        hm_actions = actions['help_menu'].actions()
        assert (hm_actions[-1].objectName() ==
                actions['new_action'].objectName())
        actions['help_menu'].removeAction(hm_actions[-1])

        # Append a new menu
        utils.addToMenu(actions['help_menu'], actions['new_menu'])
        hm_actions = actions['help_menu'].actions()
        assert (hm_actions[-1].menu().objectName() ==
                actions['new_menu'].objectName())
        actions['help_menu'].removeAction(hm_actions[-1])

    def test_addActions(self, actions):
        utils.addActions(actions['help_menu'], [None], {})
        hm_actions = actions['help_menu'].actions()
        assert hm_actions[-1].isSeparator()
        actions['help_menu'].removeAction(hm_actions[-1])

        utils.addActions(actions['help_menu'], [actions['new_menu']], {})
        hm_actions = actions['help_menu'].actions()
        assert hm_actions[-1].menu() is not None
        actions['help_menu'].removeAction(hm_actions[-1])

        utils.addActions(actions['help_menu'], ['new_action'],
                         {'new_action': actions['new_action']})
        hm_actions = actions['help_menu'].actions()
        assert hm_actions[-1].objectName() == 'testaction'
        actions['help_menu'].removeAction(hm_actions[-1])

    def test_checkFileExtension(self):
        assert utils.checkFileExtension('test') == 'test.h5'
        assert utils.checkFileExtension('test.ext') == 'test.ext'

    def test_createIcons(self):
        large_icons = frozenset(['document-close'])
        small_icons = frozenset(['document-close'])
        icons_dict = {}
        utils.createIcons(large_icons, small_icons, icons_dict)
        assert sorted(icons_dict.keys()) == ['', 'document-close',
                                             'vitables_wm']
        assert isinstance(icons_dict['document-close'], QtGui.QIcon)

    def test_forwardPath(self):
        filepath = 'C:\\Users\\my_name\\Desktop\\'
        assert utils.forwardPath(filepath) == 'C:/Users/my_name/Desktop/'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module displays in a dialog tables.Leaf information collected by the
:mod:`vitables.nodeprops.nodeinfo` module.

Users' attributes can be edited if the database has been opened in read-write
mode. Otherwise all shown information is read-only.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtGui
from qtpy import QtWidgets

from vitables.nodeprops import attrpropdlg
from vitables.nodeprops import leafproppage

translate = QtWidgets.QApplication.translate


class LeafPropDlg(attrpropdlg.AttrPropDlg):
    """
    Leaf properties dialog.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This class displays a tabbed dialog that shows some properties of
    the selected node. First tab, General, shows general properties like
    name, path, type etc. The second and third tabs show the system and
    user attributes in a tabular way.

    Beware that data types shown in the General page are `PyTables` data
    types so we can deal with `enum`, `time64` and `pseudoatoms` (none of them
    are supported by ``numpy``).
    However data types shown in the System and User attributes pages are
    ``numpy`` data types because `PyTables` attributes are stored as ``numpy``
    arrays.

    :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
      describing a given node
    """

    def __init__(self, info):
        """Setup the Properties dialog."""

        super(LeafPropDlg, self).__init__(info)

        general_page = leafproppage.LeafPropPage(info)
        self.tabw.insertTab(0, general_page, 'General')
        self.tabw.setCurrentIndex(0)

        if info.node_type == 'table':
            self.setWindowTitle(translate('LeafPropDlg', 'Table properties',
            'Dlg caption'))
        else:
            self.setWindowTitle(translate('LeafPropDlg', 'Array properties',
                'Dlg caption'))
            general_page.arrayPage()

        self.show()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"This creates an HDF5 file that contains soft and external links"

import os

import tables as tb

output_dir = '../misc'
try:
    os.mkdir(output_dir)
except OSError:
    pass

# Open a new empty HDF5 file
hdf5_name = 'links_examples.h5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
fileh = tb.open_file(filepath_hdf5, mode = "w")

# Create some groups and add datasets to them
garrays = fileh.create_group('/', 'arrays')
a1 = fileh.create_carray(garrays, 'a1', tb.Int64Atom(shape=(2,)), shape=(100, 3),
    title='A linked array')
gtables = fileh.create_group('/', 'tables')
t1 = fileh.create_table(gtables, 't1',
    {'field1': tb.FloatCol(), 'field2': tb.IntCol()}, title='A linked table')

# Create a group and put some links there
glinks = fileh.create_group('/', 'links')

# Hard link to table t1. If we remove t1 it will still be accessible via ht1.
# Hard links are not a subclass of tables.link.Link so they don't have a
# target attribute. Because hard links behave like regular nodes one can't
# infer if ht1 is a hard link by calling its __str__ method
ht1 = fileh.create_hard_link(glinks, 'ht1', '/tables/t1')
t1.remove()

# Soft link to array a1
sa1 = fileh.create_soft_link(glinks, 'sa1', '/arrays/a1')

# Soft link to table t1. This is a dangling link because it points to a
# non-existing node. In order to get the pointed node the soft links are
# callable >>> psa1 = sa1()
st1 = fileh.create_soft_link(glinks, 'st1', '/tables/t1')
# Recreate the pointed table so the link is not dangling
hht1 = fileh.create_hard_link(gtables, 't1', '/links/ht1')

# External links. They behave like soft links but point to nodes living in
# a different file
hdf5_name = 'external_file.h5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
fileh2 = tb.open_file(filepath_hdf5, mode = "w")
new_a1 = a1.copy(fileh2.root, 'a1')
fileh2.close()
sa1.remove()
ea1 = fileh.create_external_link(glinks, 'ea1', 'external_file.h5:/a1')

fileh.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a data structure to be used for the model of the databases
tree. The data structure is equivalent to a link node in a `PyTables` file.
"""

__docformat__ = 'restructuredtext'

import tables

import vitables.utils
from vitables.h5db import tlink_editor
from vitables.nodeprops import nodeinfo
from vitables.nodeprops import linkpropdlg

class LinkNode(object):
    """
    A leaf node in the tree of databases model.

    :Parameters:

    - `parent`: the parent of the node.
    - `name`: the name of the node
    """

    def __init__(self, model, parent, name):
        """Create a link node for the tree of databases model.

        A LinkNode represents a link (soft or external) of a `PyTables`
        file and has a parent (a group node of the tree of databases
        model) and a name.
        """

        self.dbt_model = model
        self.parent = parent
        self.node = parent.node._f_get_child(name)
        if isinstance(self.node(), tables.link.Link):
            self.node = self.node()

        self.has_view = False

        self.target = self.node.target
        if hasattr(self.target, 'extfile'):
            self.link_type = 'external'
        else:
            self.link_type = 'soft'

        # Attributes that the tree of databases view will use
        # name --> DisplayRole
        # nodepath --> ToolTipRole
        # as_record --> StatusTipRole
        # icon --> DecorationRole
        self.name = name
        parentpath = parent.nodepath
        if parentpath.endswith('/'):
            parentpath = parentpath[:-1]
        self.nodepath = '{0}/{1}'.format(parentpath, name)
        self.filepath = parent.filepath
        self.as_record = '{0}'.format(self.node)

        # Set the node icon
        icons = vitables.utils.getIcons()
        if isinstance(self.node(), tables.Table):
            self.node_kind = 'table'
            self.icon = icons['link_table']
        elif isinstance(self.node(), tables.VLArray):
            self.node_kind = 'vlarray'
            data_type = self.node().atom.type
            if data_type in ['vlstring', 'vlunicode']:
                self.icon = icons['vlstring']
            elif data_type == 'object':
                self.icon = icons['object']
            else:
                self.icon = icons['link_vlarray']
        elif isinstance(self.node(), tables.EArray):
            self.node_kind = 'earray'
            self.icon = icons['link_earray']
        elif isinstance(self.node(), tables.CArray):
            self.node_kind = 'carray'
            self.icon = icons['link_carray']
        elif isinstance(self.node(), tables.Array):
            self.node_kind = 'array'
            self.icon = icons['link_array']
        elif isinstance(self.node(), tables.UnImplemented):
            self.node_kind = 'image-missing'
            self.icon = icons['image-missing']


    def row(self):
        """The position of this node in the parent's list of children.
        """

        if self.parent:
            return self.parent.children.index(self)

        return 0


    def editor(self):
        """Return an instance of `TLinkEditor`.
        """
        return tlink_editor.TLinkEditor(self.dbt_model.getDBDoc(self.filepath))


    def properties(self):
        """The Properties dialog for this node.
        """

        info = nodeinfo.NodeInfo(self)
        linkpropdlg.LinkPropDlg(info)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Test class for filenode nodes."""

import os.path

import pytest

import vitables.filenodeutils as fnutils
import vitables.vttables.filenodebuffer as fnbuffer


@pytest.mark.usefixtures('launcher', 'h5file')
class TestFilenode(object):
    """Test class for modules filenodutils and filenodebuffer."""

    @pytest.fixture()
    def fnode(self, h5file):
        return h5file.get_node('/filenode')

    def test_isFilenode(self, fnode, h5file):
        # Filenode node
        assert fnutils.isFilenode(fnode)

        # No filenode node
        node = h5file.get_node('/tables/Particles')
        assert not fnutils.isFilenode(node)

    def test_filenodeToFile(self, launcher, fnode):
        file = fnutils.filenodeToFile(fnode)
        assert os.path.isfile(file)

        with open(file, 'r') as f:
            lines = f.readlines()
        assert lines[-1] == 'This is the last line.\n'

        # We have to do this by hand because the VTApp.nodeOpen method is
        # not being called
        launcher.vtapp_object.filenodes_map[fnode] = file, len(lines)

    def test_filenodeTotalRows(self, fnode):
        assert fnutils.filenodeTotalRows(fnode) == 11

    def test_fnbTotalNRows(self, fnode):
        # The filenode buffer
        fnb = fnbuffer.FilenodeBuffer(fnode)
        assert fnb.total_nrows() == 11

    def test_fnbReadBuffer(self, fnode):
        # The filenode buffer
        fnb = fnbuffer.FilenodeBuffer(fnode)
        fnb.readBuffer(0, 100)
        assert fnb.chunk[0] == ('This is a line inserted programmatically ' 
                                'at position 0\n')
        assert fnb.chunk[-1] == 'This is the last line.\n'

    def test_fnbGetCell(self, fnode):
        # The filenode buffer
        fnb = fnbuffer.FilenodeBuffer(fnode)
        fnb.readBuffer(0, 100)
        assert fnb.getCell(2, 5) == ('This is a line inserted ' 
                                     'programmatically at position 2\n')



#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
# Initial draft author: Jorge Ibanez jorge.ibannez@uam.es
#

"""Module that provides import `CSV` files into `PyTables` arrays and tables.

The pipeline for importing a `CSV` file is::

    CSV file --> numpy array --> tables.Leaf

The ``numpy`` array is created via `numpy.genfromtxt`. The `tables.Leaf`
instance is created using the appropriate constructors.

Beware that importing big files is a slow process because the whole file
has to be read from disk, transformed and write back to disk again so
there is a lot of disk IO.

Other aspects to take into account:

  - creation of `tables.Array` datasets requires the ``numpy`` array containing
    the whole `CSV` file to be loaded in memory. If the file is large enough
    you can run out of memory.

  - creation of `tables.CArray` datasets requires an additional parsing of the
    whole `CSV` file in order to find out its number of rows (it is a required
    argument of the `tables.CArray` constructor).

  - there is a penalty performance when string dtypes are involved. The reason
    is that string fields use to have variable length so, before the ``numpy``
    array is created, we need to find out the minimum itemsize required for
    storing those string fields with no lose of data. This step requires an
    additional parsing of the whole `CSV` file.

  - `CSV` files containing N-dimensional fields are always imported with `str`
    dtype. This is a limitation of `numpy.genfromtxt`.
"""

import logging
import os
import traceback

import numpy
import tables
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.csv.csvutils as csvutils
import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
TYPE_ERROR = translate(
    'ImportCSV', 'Please, make sure that you are importing a '
    'homogeneous dataset.', 'CSV file not imported error')

log = logging.getLogger(__name__)


class ImportCSV(QtCore.QObject):
    """Provides CSV import capabilities for tables and arrays.

    Some minor flaws: multidimensional fields are not well supported.
    They are imported as strings.
    """

    def __init__(self):
        """The class constructor.
        """

        super(ImportCSV, self).__init__()

        # Get a reference to the application instance
        self.vtapp = vitables.utils.getVTApp()
        if self.vtapp is None:
            return

        self.vtgui = vitables.utils.getGui()
        self.dbt_model = self.vtgui.dbs_tree_model
        self.dbt_view = self.vtgui.dbs_tree_view

        # Add an entry under the File menu
        self.icons_dictionary = vitables.utils.getIcons()
        self.addEntry()

    def addEntry(self):
        """Add the `Import CSV...` entry to the `File` menu.
        """

        self.import_csv_submenu = QtWidgets.QMenu(
            translate('ImportCSV', 'I&mport from CSV...',
                      'File -> Import CSV'))
        self.import_csv_submenu.setSeparatorsCollapsible(False)
        self.import_csv_submenu.setIcon(
            self.icons_dictionary['document-import'])
        self.import_csv_submenu.setObjectName('import_csv_submenu')

        # Create the actions
        actions = {}
        actions['import_csv_table'] = QtWidgets.QAction(
            translate('ImportCSV', "Import &Table...",
                      "Import table from CSV file"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.csv2Table,
            statusTip=translate(
                'ImportCSV',
                "Import Table from plain CSV file",
                "Status bar text for File -> Import CSV... -> Import Table"))
        actions['import_csv_table'].setObjectName('import_csv_table')

        actions['import_csv_array'] = QtWidgets.QAction(
            translate('ImportCSV', "Import &Array...",
                      "Import array from CSV file"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.csv2Array,
            statusTip=translate(
                'ImportCSV',
                "Import Array from plain CSV file",
                "Status bar text for File -> Import CSV... -> Import Array"))
        actions['import_csv_array'].setObjectName('import_csv_array')

        actions['import_csv_carray'] = QtWidgets.QAction(
            translate('ImportCSV', "Import &CArray...",
                      "Import carray from CSV file"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.csv2CArray,
            statusTip=translate(
                'ImportCSV',
                "Import CArray from plain CSV file",
                "Status bar text for File -> Import CSV... -> Import CArray"))
        actions['import_csv_carray'].setObjectName('import_csv_carray')

        actions['import_csv_earray'] = QtWidgets.QAction(
            translate('ImportCSV', "Import &EArray...",
                      "Import earray from CSV file"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.csv2EArray,
            statusTip=translate(
                'ImportCSV',
                "Import EArray from plain CSV file",
                "Status bar text for File -> Import CSV... -> Import EArray"))
        actions['import_csv_earray'].setObjectName('import_csv_earray')

        actions['import_csv_separator'] = QtWidgets.QAction(self)
        actions['import_csv_separator'].setSeparator(True)
        actions['import_csv_separator'].setObjectName('import_csv_separator')

        # Add actions to the Import submenu
        keys = ('import_csv_table', 'import_csv_array', 'import_csv_carray',
                'import_csv_earray')
        vitables.utils.addActions(self.import_csv_submenu, keys, actions)

        # Add submenu to file menu before the Close File action
        vitables.utils.insertInMenu(
            self.vtgui.file_menu, self.import_csv_submenu, 'fileClose')
        sep = actions['import_csv_separator']
        vitables.utils.insertInMenu(self.vtgui.file_menu, sep, 'fileClose')

        # Add submenu to file context menu before the Close File action
        vitables.utils.insertInMenu(self.vtgui.view_cm,
                                    self.import_csv_submenu, 'fileClose')
        vitables.utils.insertInMenu(self.vtgui.view_cm, sep, 'fileClose')

    def createDestFile(self, filepath):
        """Create the `PyTables` file where the `CSV` file will be imported.

        :Parameter filepath: the `PyTables` file filepath
        """

        dbdoc = None
        try:
            dirname, filename = os.path.split(filepath)
            root = os.path.splitext(filename)[0]
            dest_filepath = vitables.utils.forwardPath(os.path.join(dirname,
                                                                    '{0}.h5'.format(root)))
            if csvutils.isValidFilepath(dest_filepath):
                dbdoc = self.dbt_model.createDBDoc(dest_filepath)
        except:
            log.error(
                translate('ImportCSV', 'Import failed because destination '
                          'file cannot be created.',
                          'A file creation error'))
            vitables.utils.formatExceptionInfo()

        return dbdoc

    def csvFilepath(self, leaf_kind):
        """Get the filepath of the source `CSV` file.

        :Parameter leaf_kind: the kind of container where data will be stored
        """

        # Call the file selector (and, if needed, customise it)
        filepath, working_dir = vitables.utils.getFilepath(
            self.vtgui, translate(
                'ImportCSV', 'Importing CSV file into {0}',
                'Caption of the Import from CSV dialog').format(leaf_kind),
            dfilter=translate('ImportCSV', """CSV Files (*.csv);;"""
                              """All Files (*)""",
                              'Filter for the Import from CSV dialog'),
            settings={'accept_mode': QtWidgets.QFileDialog.AcceptOpen,
                      'file_mode': QtWidgets.QFileDialog.ExistingFile,
                      'history': self.vtapp.file_selector_history,
                      'label': translate('ImportCSV', 'Import',
                                         'Accept button text for QFileDialog')}
        )

        if not filepath:
            # The user has canceled the dialog
            return

        # Update the history of the file selector widget
        self.vtapp.updateFSHistory(working_dir)

        return filepath

    def updateTree(self, filepath):
        """Update the databases tree once the `CSV` file has been imported.

        When the destination h5 file is created and added to the databases tree
        it has no nodes. Once the `CSV` file has been imported into a
        `PyTables` container we update the representation of the h5 file in the
        tree so that users can see that the file has a leaf. Eventually, the
        root node of the imported file is selected so that users can locate it
        immediately.

        :Parameter filepath: the filepath of the destination h5 file
        """

        for row, child in enumerate(self.dbt_model.root.children):
            if child.filepath == filepath:
                index = self.dbt_model.index(row, 0, QtCore.QModelIndex())
                self.dbt_model.lazyAddChildren(index)
                self.dbt_view.setCurrentIndex(index)

    def csv2Table(self):
        """Import a plain `CSV` file into a `tables.Array` object.
        """

        kind = 'Table'
        filepath = self.csvFilepath(kind)
        if filepath is None:
            return

        # Import the CSV content
        try:
            QtWidgets.qApp.processEvents()
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            input_handler = open(filepath, 'r+')
            try:
                (nrows, descr, has_header) = csvutils.tableInfo(input_handler)
            except Exception as inst:
                print(traceback.format_exc())

            # Create the dataset
            dbdoc = self.createDestFile(filepath)
            if dbdoc is None:
                return
            io_filters = tables.Filters(complevel=9, complib='lzo')
            dataset_name = "imported_{0}".format(kind)
            atitle = \
                'Source CSV file {0}'.format(os.path.basename(filepath))
            dataset = dbdoc.h5file.create_table(
                '/', dataset_name, descr, title=atitle, filters=io_filters,
                expectedrows=nrows)
            # Fill the dataset in a memory efficient way
            input_handler.seek(0)
            if has_header:
                # Skip the header line
                input_handler.readline()
            chunk_size = 10000
            buf_size = chunk_size * dataset.rowsize
            read_fh = input_handler.readlines
            buf = read_fh(buf_size)
            while buf:
                idata = csvutils.getArray(buf)
                # Append data to the dataset
                dataset.append(idata)
                dataset.flush()
                del idata
                buf = read_fh(buf_size)
            dbdoc.h5file.flush()
            self.updateTree(dbdoc.filepath)
        except:
            vitables.utils.formatExceptionInfo()
        finally:
            QtWidgets.qApp.restoreOverrideCursor()
            input_handler.close()

    def csv2EArray(self):
        """Import a plain `CSV` file into a `tables.EArray` object.

        This is a slot method. See :meth:`addEntry` method for details.
        """

        kind = 'EArray'
        filepath = self.csvFilepath(kind)
        if filepath is None:
            return

        # Import the CSV content
        try:
            QtWidgets.qApp.processEvents()
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            chunk_size = 10000
            input_handler = open(filepath, 'r+')
            (nrows, atom, array_shape) = csvutils.earrayInfo(input_handler)

            # Create the dataset
            dbdoc = self.createDestFile(filepath)
            if dbdoc is None:
                return
            io_filters = tables.Filters(complevel=9, complib='lzo')
            dataset_name = "imported_{0}".format(kind)
            atitle = 'Source CSV file {0}'.format(os.path.basename(filepath))
            dataset = dbdoc.h5file.create_earray(
                '/', dataset_name, atom, array_shape, title=atitle,
                filters=io_filters, expectedrows=nrows)

            # Fill the dataset in a memory effcient way
            input_handler.seek(0)
            chunk_size = 10000
            buf_size = chunk_size * dataset.rowsize
            read_fh = input_handler.readlines
            buf = read_fh(buf_size)
            while buf:
                idata = csvutils.getArray(buf)
                # Append data to the dataset
                dataset.append(idata)
                dataset.flush()
                del idata
                buf = read_fh(buf_size)
            dbdoc.h5file.flush()
            self.updateTree(dbdoc.filepath)
        except ValueError:
            log.error(TYPE_ERROR)
        except:
            vitables.utils.formatExceptionInfo()
        finally:
            QtWidgets.qApp.restoreOverrideCursor()
            input_handler.close()

    def csv2CArray(self):
        """Import a plain `CSV` file into a `tables.CArray` object.

        This is a slot method. See :meth:`addEntry` method for details.
        """

        kind = 'CArray'
        filepath = self.csvFilepath(kind)
        if filepath is None:
            return

        # Import the CSV content
        try:
            QtWidgets.qApp.processEvents()
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            chunk_size = 10000
            input_handler = open(filepath, 'r+')
            (atom, array_shape) = csvutils.carrayInfo(input_handler)

            # Create the dataset
            dbdoc = self.createDestFile(filepath)
            if dbdoc is None:
                return
            io_filters = tables.Filters(complevel=9, complib='lzo')
            dataset_name = "imported_{0}".format(kind)
            atitle = 'Source CSV file {0}'.format(os.path.basename(filepath))
            dataset = dbdoc.h5file.create_carray(
                '/', dataset_name, atom, array_shape, title=atitle,
                filters=io_filters)

            # Fill the dataset in a memory effcient way
            input_handler.seek(0)
            chunk_size = 10000
            buf_size = chunk_size * dataset.rowsize
            read_fh = input_handler.readlines
            buf = read_fh(buf_size)
            start = 0
            while buf:
                idata = csvutils.getArray(buf)
                stop = start + idata.shape[0]
                # Append data to the dataset
                dataset[start:stop] = idata
                dataset.flush()
                del idata
                start = stop
                buf = read_fh(buf_size)
            dbdoc.h5file.flush()
            self.updateTree(dbdoc.filepath)
        except ValueError:
            log.error(TYPE_ERROR)
        except:
            vitables.utils.formatExceptionInfo()
        finally:
            QtWidgets.qApp.restoreOverrideCursor()
            input_handler.close()

    def csv2Array(self):
        """Import a plain `CSV` file into a `tables.Array` object.

        This is a slot method. See :meth:`addEntry` method for details.
        """

        kind = 'Array'
        filepath = self.csvFilepath(kind)
        if filepath is None:
            return

        # Import the CSV content
        try:
            QtWidgets.qApp.processEvents()
            QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
            # The dtypes are determined by the contents of each column
            # Multidimensional columns will have string datatype
            data = numpy.genfromtxt(filepath, delimiter=',', dtype=None)
        except TypeError:
            data = None
            dbdoc = None
            log.error(TYPE_ERROR)
        else:
            try:
                # Create the array
                dbdoc = self.createDestFile(filepath)
                if dbdoc is None:
                    return
                array_name = "imported_{0}".format(kind)
                title = 'Imported from CSV file {0}'.\
                    format(os.path.basename(filepath))
                dbdoc.h5file.create_array('/', array_name, data, title=title)
                dbdoc.h5file.flush()
                self.updateTree(dbdoc.filepath)
            except TypeError:
                log.error(TYPE_ERROR)
            except tables.NodeError:
                vitables.utils.formatExceptionInfo()
        finally:
            del data
            QtWidgets.qApp.restoreOverrideCursor()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines a view for the tree of databases model.

This view is used to display the tree of open databases. Each top level
node of the tree contains the object tree of a `PyTables`/`HDF5` database.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

from vitables.h5db.nodeitemdelegate import NodeItemDelegate


translate = QtWidgets.QApplication.translate


class DBsTreeView(QtWidgets.QTreeView):
    """
    A view for the databases tree model.

    :Parameters:
        - `vtapp`: the VTAPP instance
        - `model`: the model for this view
    """


    dbsTreeViewCreated = QtCore.Signal(QtWidgets.QTreeView)

    def __init__(self, vtapp, vtgui, model, parent=None):
        """Create the view.
        """

        super(DBsTreeView, self).__init__(parent)

        # The model
        self.setModel(model)
        self.dbt_model = model
        self.smodel = self.selectionModel()

        self.vtapp = vtapp
        self.vtgui = vtgui

        # The custom delegate used for editing items
        self.setItemDelegate(NodeItemDelegate(vtgui, self))
        self.setObjectName('dbs_tree_view')

        # The frame specification
        self.frame_style = {'shape': self.frameShape(),
            'shadow': self.frameShadow(),
            'lwidth': self.lineWidth(),
            'foreground': self.palette().color(QtGui.QPalette.Active,
                QtGui.QPalette.WindowText)}

        # Setup drag and drop
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)

        # Misc. setup
        self.setRootIsDecorated(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        # Whether selections are done in terms of single items, rows or columns
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        # Whether the user can select one or many items
        # Changed from SingleSelection to ExtendedSelection in commit 403a4c3
        # but I don't know why. I revert it to SingleSelection or a
        # tables.ClosedNodeError is randomly raised when a group is moved to a
        # different file
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setWhatsThis(translate('DBsTreeView',
            """<qt>
            <h3>The Tree of databases</h3>
            For every open database this widget shows the object tree,
            a graphical representation<br>of the data hierarchy stored
            in the database.</qt>""",
            'WhatsThis help for the tree pane'))

        # Connect signals to slots
        self.customContextMenuRequested.connect(self.createCustomContextMenu)
        self.activated.connect(self.activateNode)
        self.expanded.connect(self.updateExpandedGroup)
        self.collapsed.connect(self.updateCollapsedGroup)
        self.dbt_model.layoutChanged.connect(self.updateColumnWidth)


    def updateColumnWidth(self):
        """Make sure that a horizontal scrollbar is shown as needed.

        This is a subtle method. As the tree view has only 1 column its
        width and the width of the viewport are always the same so the
        horizontal scrollbar is never shown. As the contents width
        changes every time the layout changes (rows are inserted or
        deleted) by resizing column to contents when it happens we
        ensure that the column and the viewport will have different
        width and the scrollbar will indeed be added as needed.
        """
        self.resizeColumnToContents(0)


    def activateNode(self, index):
        """Expand an item via `Enter`/`Return` key or mouse double click.

        When the user activates a collapsed item (by pressing `Enter`, `Return`
        or by double clicking the item) then it is expanded. If the user
        activates the node by double clicking on it while the `Shift` key is
        pressed, the item is edited (if editing is enabled).

        Lazy population of the model is partially implemented in this
        method. Expanded items are updated so that children items are added if
        needed. This fact improves enormously the performance when files
        whit a large number of nodes are opened.

        This method is a slot connected to the activated(QModelIndex) signal
        in the ctor.

        :Parameter index: the index of the activated item
        """

        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if (modifiers & QtCore.Qt.ShiftModifier) or \
                (modifiers & QtCore.Qt.ControlModifier):
            return
        node = self.dbt_model.nodeFromIndex(index)
        if node.node_kind.count('group'):
            if not self.isExpanded(index):
                self.expand(index)
        elif node.has_view:
            ## Activate already-open window.
            #
            wrkspc = self.vtgui.workspace
            pcurrent = QtCore.QPersistentModelIndex(index)
            for window in wrkspc .subWindowList():
                if pcurrent == window.pindex:
                    wrkspc.setActiveSubWindow(window)
        else:
            self.vtapp.nodeOpen(index)


    def updateCollapsedGroup(self, index):
        """After collapsing a group update its icon.

        This method is a slot connected to the `collapsed(QModelIndex)` signal
        in the ctor.

        :Parameter index: the index of the collapsed group
        """

        node = self.dbt_model.nodeFromIndex(index)
        if node.node_kind == 'group':
            self.dbt_model.setData(index, node.closed_folder,
                QtCore.Qt.DecorationRole)
        self.smodel.clearSelection()
        self.smodel.setCurrentIndex(index,
            QtCore.QItemSelectionModel.SelectCurrent)
        self.update(index)


    def updateExpandedGroup(self, index):
        """After a group expansion, update the icon and the displayed children.

        Lazy population of the model is partially implemented in this
        method. Expanded items are updated so that children items are added if
        needed. This fact reduces enormously the opening times for files
        whit a large number of nodes and also saves memory.

        This method is a slot connected to the `expanded(QModelIndex)` signal
        in the ctor.

        :Parameter index: the index of the expanded item
        """

        node = self.dbt_model.nodeFromIndex(index)
        node_kind = node.node_kind
        if node_kind == 'group':
            self.dbt_model.setData(index, node.open_folder,
                QtCore.Qt.DecorationRole)
        if node_kind in ['group', 'root group']:
            if not node.updated:
                self.dbt_model.lazyAddChildren(index)
                node.updated = True
                self.smodel.clearSelection()
                self.smodel.setCurrentIndex(index,
                    QtCore.QItemSelectionModel.SelectCurrent)


    def createCustomContextMenu(self, pos):
        """
        A context menu for the tree of databases view.

        When an item of the tree view is right clicked, a context popup
        menu is displayed. The content of the popup depends on the
        clicked element.

        :Parameter pos: the local position at which the menu will popup
        """

        index = self.indexAt(pos)
        if not index.isValid():
            kind = 'view'
        else:
            node = self.dbt_model.nodeFromIndex(index)
            kind = node.node_kind
        pos = self.mapToGlobal(pos)
        self.vtgui.popupContextMenu(kind, pos)


    def selectNode(self, index):
        """Select the given index.

        :Parameter `index`: the model index being selected
        """

        self.smodel.clearSelection()
        self.smodel.setCurrentIndex(index,
            QtCore.QItemSelectionModel.SelectCurrent)


    def mouseDoubleClickEvent(self, event):
        """Specialised handler for mouse double click events.

        When a node is double clicked in the tree of databases pane:

        - if the node can be renamed and the `Shift` key is pressed then
          rename the node
        - if the node is a leaf with no view and the `Shift` key is not pressed
          then open the node
        - if the node is a collpased group and the `Shift` key is not pressed
          then expand the group

        :Parameter event: the event being processed

        """

        modifier = event.modifiers()
        current = self.currentIndex()
        if modifier == QtCore.Qt.ShiftModifier:
            if current.flags() & QtCore.Qt.ItemIsEditable:
                self.edit(current)
        else:
            self.activateNode(current)


    def dragMoveEvent(self, event):
        """
        Event handler for `QDragMoveEvent` events.

        Dragging icons from the Desktop/Files Manager into the tree view is
        supported.

        :Parameter event: the event being processed.
        """

        # The dragged object (one or more files) has a MIME type (namely
        # text/uri-list) and some associated data (the list of URLs of the
        # dragged files). All that information can be accessed via the event
        # MimeData() method.
        # The widget should examine that information and accept the drop if
        # appropriate.
        mime_data = event.mimeData()
        if mime_data.hasFormat('text/uri-list'):
            event.setDropAction(QtCore.Qt.CopyAction)
            event.acceptProposedAction()
        else:
            return QtWidgets.QTreeView.dragMoveEvent(self, event)


    def dropEvent(self, event):
        """
        Event handler for `QDropEvent` events.

        This event is sent when a drag and drop action is completed. In our case
        if an icon is dropped on the tree view then the icon URL is converted to
        a path (which we assume to be an `HDF5` file path) and ``ViTables``
        tries to open it.

        :Parameter event: the event being processed.
        """

        # The dropped object (one or more files) has a MIME type (namely
        # text/uri-list) and some associated data (the list of URLs of the
        # dropped files). All that information can be accessed via the event
        # MimeData() method.
        # The widget should examine that information and launch the required
        # actions.
        mime_data = event.mimeData()
        if mime_data.hasFormat('text/uri-list'):
            if self.dbt_model.dropMimeData(
                mime_data, QtCore.Qt.CopyAction, -1, -1, self.currentIndex()):
                event.setDropAction(QtCore.Qt.CopyAction)
                event.accept()
                self.setFocus(True)
        else:
            QtWidgets.QTreeView.dropEvent(self, event)


    def focusInEvent(self, event):
        """Specialised handler for focus events.

        Repaint differently the databases tree view frame when it gets the
        keyboard focus so that users can realize easily about this focus
        change.

        :Parameter event: the event being processed
        """

        self.setLineWidth(2)
        self.setFrameStyle(QtWidgets.QFrame.Panel|QtWidgets.QFrame.Plain)
        pal = self.palette()
        pal.setColor(QtGui.QPalette.Active, QtGui.QPalette.WindowText,
            QtCore.Qt.darkBlue)
        QtWidgets.QTreeView.focusInEvent(self, event)


    def focusOutEvent(self, event):
        """Specialised handler for focus events.

        Repaint differently the databases tree view frame when it looses the
        keyboard focus so that users can realize easily about this focus
        change.

        :Parameter event: the event being processed
        """

        self.setLineWidth(self.frame_style['lwidth'])
        self.setFrameShape(self.frame_style['shape'])
        self.setFrameShadow(self.frame_style['shadow'])
        pal = self.palette()
        pal.setColor(QtGui.QPalette.Active, QtGui.QPalette.WindowText,
            self.frame_style['foreground'])
        QtWidgets.QTreeView.focusOutEvent(self, event)


if __name__ == '__main__':
    import sys
    APP = QtWidgets.QApplication(sys.argv)
    TREE = DBsTreeView()
    TREE.show()
    APP.exec_()



import sip
sip.setapi('QString', 2)
sip.setapi('QVariant', 2)

import nose.tools as nt

import vitables.calculator.calculator as vtc


class TestCalculator:
    """Test calculator functionality."""

    def test_extract_identifiers(self):
        """Test function that extract marked identifiers from an expression."""
        empty_set = vtc.extract_identifiers('')
        nt.assert_true(len(empty_set) == 0)
        nt.assert_equal(set(['$data']), vtc.extract_identifiers('$data'))
        nt.assert_equal(set(['$data']), vtc.extract_identifiers('+$data*'))
        nt.assert_equal(set(['$data']),
                        vtc.extract_identifiers('+$data*$data'))
        nt.assert_equal(set(['$data', '$data1']),
                        vtc.extract_identifiers('+$data*$data1 + test'))
        nt.assert_equal(set(['$data', '$test.data1']),
                        vtc.extract_identifiers('+$data*$test.data1 + test'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
A custom delegate for rendering selected cells.

As explained in the leaf_view module the current cell must be rendered
differently depending on the buffer being displayed. If the dataset row of the
current/selected cell is in the range of rows being displayed then the cell
should be rendered in the standard way. If it is not then it should always be
rendered unselected.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

class LeafDelegate(QtWidgets.QStyledItemDelegate):
    """
    A delegate for rendering selected cells.

    :Parameter parent: the parent of this widget
    """


    def paint(self, painter, option, index):
        """Renders the delegate for the item specified by index.

        This method handles specially the result returned by the model data()
        method for the Qt.BackgroundRole role. Typically, if the cell being
        rendered is selected then the data() returned value is ignored and the
        value set by the desktop (KDE, Gnome...) is used. We need to change
        that behavior as explained in the module docstring.

        The following properties of the style option are used for customising
        the painting: state (which holds the state flags), rect (which holds
        the area that should be used for painting) and palette (which holds the
        palette that should be used when painting)

        :Parameters:

        - `painter`: the painter used for rendering
        - `option`: the style option used for rendering
        - `index`: the index of the rendered item
        """

        # option.state is an ORed combination of flags
        if (option.state & QtWidgets.QStyle.State_Selected):
            model = index.model()
            buffer_start = model.start
            cell = index.model().selected_cell
            if ((index == cell['index']) and \
                    (buffer_start != cell['buffer_start'])):
                painter.save()
                self.initStyleOption(option, index)
                background = option.palette.color(QtGui.QPalette.Base)
                foreground = option.palette.color(QtGui.QPalette.Text)
                painter.setBrush(QtGui.QBrush(background))
                painter.fillRect(option.rect, painter.brush())
                painter.translate(option.rect.x() + 3, option.rect.y())
                painter.setBrush(QtGui.QBrush(foreground))
                painter.drawText(
                    option.rect,
                    QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop,
                    model.data(index))
                painter.restore()
            else:
                QtWidgets.QStyledItemDelegate.paint(self, painter, option, index)
        else:
            QtWidgets.QStyledItemDelegate.paint(self, painter, option, index)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module implements a buffer used to access the real data contained in
`PyTables` filenode.

By using this buffer we speed up the access to the stored data. As a
consequence, views (widgets showing a tabular representation of the dataset)
are painted much faster too.
"""

import logging
import warnings
import linecache

import numpy
import tables

from .. import utils as vtutils

__docformat__ = 'restructuredtext'

# Restrict the available flavors to 'numpy' so that reading a leaf
# always return a numpy array instead of an object of the kind indicated
# by the leaf flavor. For VLArrays the read data is returned as a list whose
# elements will be numpy arrays.
tables.restrict_flavors(keep=['numpy'])
warnings.filterwarnings('ignore', category=tables.FlavorWarning)
warnings.filterwarnings('ignore', category=tables.NaturalNameWarning)

log = logging.getLogger(__name__)


class FilenodeBuffer(object):
    """Buffer used to access the real data contained in `PyTables` filenode.

    Note that the buffer number of rows **must** be at least equal to
    the number of rows of the table widget it is going to fill. This
    way we avoid to have partially filled tables.  Also note that rows
    in buffer are numbered from 0 to N (as it happens with the data
    source).

    Filenodes are displayed in Mx1 table widgets.

    :Parameter leaf:
        the data source (`tables.EArray` instance) from which data are
        going to be read.
    """

    def __init__(self, leaf):
        """
        Initializes the buffer.
        """

        self.leaf = leaf
        # The structure where read data will be stored.
        self.chunk = numpy.array([])

        vtapp = vtutils.getApp()
        self.temp_filenode, self.total_rows = vtapp.filenodes_map[leaf]

    def __del__(self):
        """Release resources before destroying the buffer.
        """
        # FIXME: PY3.5+ leaks resources (use finalizer instead).
        self.chunk = None

    def total_nrows(self):
        return self.total_rows

    def readBuffer(self, start, stop):
        """
        Read the whole filenode in memory using the linecache module.
        Data is returned as a Python list of byte strings.

        :Parameters:
        :param start: the document row that is the first row of the chunk.
        :param stop: the last row to read, inclusive.
        """

        # The line numbers of files read by linecache start with 1
        if start == 0:
            start += 1
        if stop > self.total_rows:
            stop = self.total_rows

        data = []
        counter = start
        while counter <= stop:
            data.append(linecache.getline(self.temp_filenode, counter))
            counter += 1
        self.chunk = data

    def getCell(self, row, col):
        """
        Returns a cell of a 1D-array view.

        :Parameters:
        - `row`: the row to which the cell belongs.
        - `col`: the column to wich the cell belongs
        :Returns: the cell at position `(row, col)` of the document
        """

        # The row-coordinate must be shifted by model.start units in order to
        # get the right chunk element.
        # chunk = [row0, row1, row2, ..., rowN]
        # and columns can be read from a given row using indexing notation
        return self.chunk[row]



#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org


"""Plugin that provides columnar organizations for arrays

The plugin let the users to mark several array views (all of them with the
same number of rows) and then display those arrays using a unique widget, one
array next to other. One can think that a set of arrays are displayed as a
table. This seems quite strange to me but is a feature request so I assume
that the plugin will be useful to some users.
"""

__docformat__ = 'restructuredtext'
__version__ = '1.1'
plugin_name = 'Columnar organization of arrays'
comment = ('Rearrange several arrays with the same number of rows and '
           'displays them in a unique widget')

import os.path

import tables

from qtpy import QtGui
from qtpy import QtCore
from qtpy import QtWidgets

import vitables
from vitables.vttables.datasheet import DataSheet
from vitables.plugins.aboutpage import AboutPage

translate = QtWidgets.QApplication.translate

_PLUGIN_FOLDER = os.path.join(os.path.dirname(__file__))


class ArrayColsOrganizer(QtCore.QObject):
    """The class which defines the plugin for columnar organization of arrays.

    It is connected to the core program so that ever time a new view
    is created this class customizes the view if it is required. In
    addition the class triggers the update of the Node menu and
    includes a `helpAbout`method that is used in the Preferences
    dialog to show information about the plugin.

    """

    UID = __name__
    NAME = plugin_name
    COMMENT = comment

    def __init__(self, parent=None):
        """Class constructor.

        Dynamically finds new instances of
        :meth:`vitables.vttables.leaf_model.LeafModel` and customizes them if
        they are arrays that can be grouped in a unique view.
        """

        super(ArrayColsOrganizer, self).__init__(parent)
        self.vtapp = vitables.utils.getVTApp()

        # An specialized object will deal with the Node menu changes
        self.menu_updater = MenuUpdater()
        self.menu_updater.addEntry()

        # Connect convenience signal defined in the vtapp module to slot
        self.vtapp.leaf_model_created.connect(self.customizeView)

    def customizeView(self, datasheet):
        """Add a checkbox in the bottom right corner of array views.

        Customization is twofold: add a checkbox to the LeafView and add a
        is_checked` attr to the view. In addition, the checkbox has a
        `source_array` attribute that points to the view containing it.

        In a regular array, if checkbox ticked, it means that the array is
        candidate to be added to other single/grouped arrays.
        In a group of arrays, if checkbox is ticked it means that the group
        can be group with other ticked single/grouped arrays. If the grouped
        array is the only ticked view then it means that it is a candidate for
        being ungrouped.

        This is a slot called every time a new view is created.

        """

        # The tables.Node instance tied to that data structure
        pt_node = datasheet.dbt_leaf.node
        if hasattr(pt_node, 'target'):
            # The selected item is a link and must be dereferenced
            leaf = pt_node()
        else:
            leaf = pt_node

        if not isinstance(leaf, tables.Table):
            cb = QtWidgets.QCheckBox(datasheet)
            cb.setToolTip(translate('ArrayColsOrganizer',
                                    "Include this Array in the 'columnar Group'",
                                    'Checkbox tooltip text'))
            datasheet.leaf_view.setCornerWidget(cb)
            # Attributes for retrieving the state of the checkbox from the
            # datasheet being customized
            cb.source_array = datasheet
            datasheet.is_checked = \
                datasheet.leaf_view.cornerWidget().checkState()

            # Connect signals to slots
            cb.stateChanged.connect(self.cbStateChanged)

    def cbStateChanged(self, state):
        """Update the `is_checked` attribute of DataSheets and GroupedArrays.
        """

        view = self.sender().source_array
        view.is_checked = state

    def helpAbout(self, parent):
        """Full description of the plugin.

        This is a convenience method which works as expected by
        :meth:preferences.preferences.Preferences.aboutPluginPage i.e.
        build a page which contains the full description of the plugin
        and, optionally, allows for its configuration.

        :Parameter about_page: the container widget for the page
        """

        # Plugin full description
        desc = {'version': __version__,
                'module_name': os.path.join(os.path.basename(__file__)),
                'folder': os.path.join(os.path.dirname(__file__)),
                'author': 'Vicent Mas <vmas@vitables.org>',
                'comment': translate('ArrayColsOrganizer',
                                     """
                    <qt><p>Plugin that provides an alternative view for
                    arrays with the same number of rows. <p>
                    The user has to open all the arrays she whishes to see in
                    a unique viewer, and tick their bottom-right corner
                    checkbox.
                    She may then click on the `Node -> Group Arrays`
                    menu entry and these arrays will be grouped
                    in a single sub-window.
                    <p>Note that all ticked arrays must have the same
                    dimensions for the `Group` menu item to enable.</qt>
                    """,
                                     'Text of an About plugin message box')}
        about_page = AboutPage(desc, parent)
        return about_page


class MenuUpdater(QtCore.QObject):
    """Specialized class for modifying the Node menu.
    """

    def __init__(self, parent=None):
        """The constructor method.
        """

        super(MenuUpdater, self).__init__(parent)
        self.vtgui = vitables.utils.getGui()

        # Connect signals to slots
        self.vtgui.node_menu.aboutToShow.connect(self.updateNodeMenu)
        self.vtgui.leaf_node_cm.aboutToShow.connect(self.updateNodeMenu)

    def addEntry(self):
        """Add the `Group Arrays` and `Ungroup Arrays` entries to `Node` menu.
        """

        object_group_icon = QtGui.QIcon()
        pixmap = QtGui.QPixmap(os.path.join(_PLUGIN_FOLDER,
                                            'icons/object-group.png'))
        object_group_icon.addPixmap(pixmap,
                                    QtGui.QIcon.Normal,
                                    QtGui.QIcon.On)

        self.group_action = QtWidgets.QAction(
            translate('MenuUpdater',
                      "&Group Arrays",
                      "Group separated arrays with the same number of rows"),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.groupArrays,
            icon=object_group_icon,
            statusTip=translate('MenuUpdater',
                                """
                Use a unique widget to display Arrays as if
                they where columns of a Table""",
                                "Status bar text for the Node -> Group Arrays action"))

        object_ungroup_icon = QtGui.QIcon()
        pixmap = QtGui.QPixmap(os.path.join(_PLUGIN_FOLDER,
                                            'icons/object-ungroup.png'))
        object_ungroup_icon.addPixmap(pixmap,
                                      QtGui.QIcon.Normal,
                                      QtGui.QIcon.On)

        self.ungroup_action = QtWidgets.QAction(
            translate('MenuUpdater',
                      "&Ungroup Arrays",
                      "Ungroup previously grouped arrays."),
            self,
            shortcut=QtGui.QKeySequence.UnknownKey,
            triggered=self.ungroupArrays,
            icon=object_ungroup_icon,
            statusTip=translate('MenuUpdater',
                                """Ungroup previously grouped arrays.""",
                                "Status bar text for the Node -> Ungroup Arrays action"))

        # Add the actions to the Node menu
        vitables.utils.addToMenu(self.vtgui.node_menu, (self.group_action,
                                                        self.ungroup_action))

        # Add the actions to the leaf context menu
        vitables.utils.addToLeafContextMenu(
            (self.group_action, self.ungroup_action))

    def updateNodeMenu(self):
        """Update (un)group_action QActions if the Node menu is about to show.

        The `group_action` is disabled when there are less than two Arrays
        checked, and also when not all checked arrays have the same number of
        rows.

        The `ungroup_action` is enabled only if there are one array group
        selected.

        """
        self.group_action.setEnabled(False)
        self.ungroup_action.setEnabled(False)
        checked_views = []
        numrows = []
        for view in self.vtgui.workspace.subWindowList():
            # Both Datasheet and GroupedArray instances have an is_checked attr
            # but Table instances don't have it.
            if hasattr(view, 'is_checked') and (view.is_checked ==
                                                QtCore.Qt.Checked):
                checked_views.append(view)
                if isinstance(view, GroupedArrays):
                    numrows.append(view.grouped_arrays_nrows)
                elif isinstance(view, DataSheet):
                    numrows.append(view.leaf_model.numrows[()])

        self.checked_views = checked_views

        if len(checked_views) == 1:
            if isinstance(checked_views[0], GroupedArrays):
                self.ungroup_action.setEnabled(True)
        elif len(checked_views) > 1:
            if numrows.count(numrows[0]) == len(numrows):
                self.group_action.setEnabled(True)
            else:
                print(translate('MenuUpdater',
                                """\nError: grouping arrays, operation cancelled. """
                                """Not all the selected arrays have the same number """
                                """of rows.""",
                                'An error message in the ArraysColsOrganizer plugin'))

    def groupArrays(self):
        """Group a set of individual arrays into the same view widget.
        """
        GroupedArrays(views=self.checked_views)

    def ungroupArrays(self):
        """Break-up a grouped array into a set of views (one per component).
        """
        GroupedArrays.ungroupArrays(self.checked_views[0])


class GroupedArrays(QtWidgets.QMdiSubWindow):
    """A widget that contains a group of Arrays.
    """

    def __init__(self, parent=None, flags=QtCore.Qt.SubWindow,
                 views=None):
        """The class constructor.
        """

        self.vtgui = vitables.utils.getGui()
        super(GroupedArrays, self).__init__(self.vtgui.workspace, flags)

        # The components from which the GroupedArrays will be made.
        # We use sets instead of lists because there seems to be a
        # problem with my_list.pop(i) when i != -1
        self.arrays = set(views)
        self.already_grouped = set([])
        for i in self.arrays.copy():
            if isinstance(i, GroupedArrays):
                self.already_grouped.add(i)
                self.arrays.remove(i)
        self.arrays = list(self.arrays)
        self.already_grouped = list(self.already_grouped)
        self.pindex = QtCore.QModelIndex()
        # Create a GroupedArrays object from the set of regular arrays
        if (len(self.arrays) > 1):
            self.combineArrays()
        # Combine regular arrays with GroupedArrays
        elif (len(self.arrays) == 1) and len(self.already_grouped):
            self.combineArrayAndGroupedArrays()
        # Combine GroupedArrays (no regular arrays in the way)
        elif len(self.already_grouped) > 1:
            self.combineGroupedArrays()

    # The following functions define the behavior of GroupedArrays objects:
    # - how they are created from several Datasheet objects
    # - how a single Datasheet object can be added to a GroupedArray object
    # - how they combine together to create a bigger GroupedArrays object
    # - how they are customized
    # - how they are closed
    # - how they are ungrouped

    def addVerticalLayout(self, container_layout, datasheet, title=None):
        """Add a vertical layout to the horizontal layout of a GroupArrays.

        This is a helper method.

        :Parameter:
        - `container_layout`: the horizontal layout of the ArraysGroup
        - `datasheet`: the datasheet being added
        """

        if not title:
            title = QtWidgets.QLabel(datasheet.windowTitle())
        datasheet.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        vertical_layout = QtWidgets.QVBoxLayout()
        vertical_layout.addWidget(title)
        vertical_layout.addWidget(datasheet)
        container_layout.addLayout(vertical_layout)

    def convenienceAttrs(self, datasheet):
        """Helper method for setting some attributes
        """

        self.is_checked = QtCore.Qt.Checked
        nrows = datasheet.leaf_model.numrows[()]
        self.grouped_arrays_nrows = nrows

    def combineArrays(self):
        """Convert a set of array views into a GroupedArray view.
        """

        self.pindex = QtCore.QModelIndex()
        # The widget of this MdiSubWindow
        self.container = QtWidgets.QWidget()

        # Fill the widget with Datasheets arranged in the container layout
        container_layout = QtWidgets.QHBoxLayout()
        while self.arrays:
            # Extract the first datasheet and manipulate it conveniently:
            # get the title, remove the frame and put both of them in a
            # vertical layout
            datasheet = self.arrays.pop()
            self.addVerticalLayout(container_layout, datasheet)
        self.container.setLayout(container_layout)

        # Some convenient attributes
        self.convenienceAttrs(datasheet)

        # Create and customize the MdiSubWindow
        self.setWidget(self.container)
        self.show()
        self.customizeGroupedViews()

    def combineArrayAndGroupedArrays(self):
        """Add an Array to this GroupedArray instance.
        """

        datasheet = self.arrays.pop()
        # Overwrite self with the last GroupedArray instance of the list
        # of grouped arrays but preserve the list of grouped arrays seen
        # by the self object. Calling self.show() won't be required because
        # the GroupedArray is already visible
        self.deleteLater()
        self.pindex = QtCore.QModelIndex()
        grouped_arrays = self.already_grouped[:]
        self = self.already_grouped.pop()
        self.already_grouped = grouped_arrays[:-1]
        self.container = self.widget()
        container_layout = self.container.layout()
        # The datasheet is manipulated conveniently:
        # get the title, remove the frame and put both of them in a
        # vertical layout
        self.addVerticalLayout(container_layout, datasheet)

        # Some convenient attributes
        self.convenienceAttrs(datasheet)

        self.customizeGroupedViews()

    def combineGroupedArrays(self):
        """Convert a set of GroupedArrays into a unique GroupedArrays view.

        :Parameter grouped_arrays: a set of GroupedArrays views
        """

        # Overwrite self with the last GroupedArray instance of the list
        # of grouped arrays but preserve the list of grouped arrays seen
        # by the self object. Calling self.show() won't be required because
        # the GroupedArray is already visible
        self.deleteLater()
        self.pindex = QtCore.QModelIndex()
        grouped_arrays = self.already_grouped[:]
        self = self.already_grouped.pop()
        self.already_grouped = grouped_arrays[:-1]
        self.container = self.widget()
        container_layout = self.container.layout()

        while self.already_grouped:
            ga = self.already_grouped.pop(-1)
            hlayout = ga.widget().layout()
            first_datasheet = hlayout.itemAt(0).itemAt(1).widget()
            for i in range(hlayout.count()):
                vl = hlayout.itemAt(i)
                title = vl.itemAt(0).widget()
                datasheet = vl.itemAt(1).widget()
                self.addVerticalLayout(container_layout, datasheet, title)
            ga.deleteLater()
#        self.repaint()

        # Some convenient attributes
        self.convenienceAttrs(first_datasheet)

        self.customizeGroupedViews()

    def customizeGroupedViews(self):
        """Make the grouped view widget usable.

        - All but the leftmost horizontal header views are hidden
        - All but the rightmost verticals scrollbars are hidden
          and connected between them

        Beware that horizontal scrollbars cannot be connected easily if the
        number of columns differ for the grouped arrays so we recommend to
        group only arrays with the same number of columns.

        :Parameter sw: the viw (QMdiSubWindow instance) being customized
        """

        # This instance will be the first GroupedArray instance of the set
        # Calling self.show() won't be required because the GroupedArray is
        # already visible
        self.container = self.widget()
        container_layout = self.container.layout()

        # The items contained in the container layout are QVBoxLAyout layout
        # boxes. These vertical layouts can contain labels and datasheets:
        # horizontal layout -> vertical layout -> widget item -> label
        #                                     |-> widget item -> datasheet
        datasheets = []
        for i in range(container_layout.count()):
            vl = container_layout.itemAt(i)
            datasheets.append(vl.itemAt(1).widget())

        nd = len(datasheets)
        for i in range(nd):
            datasheets[nd - 1].widget().verticalScrollBar().valueChanged.connect(
                datasheets[i].widget().verticalScrollBar().setValue)
            if i < (nd - 1):
                datasheets[i].widget().verticalScrollBar().hide()
                datasheets[i].widget().setCornerWidget(None)
                datasheets[i + 1].widget().verticalHeader().hide()

    def closeEvent(self, event):
        """ Propagate the close event.
        In the process, self.widget().closeEvent will be called
        """

        sw_layout = self.widget().layout()
        vl_count = sw_layout.count()
        for i in range(vl_count):
            layout = sw_layout.itemAt(i)
            datasheet = layout.itemAt(1).widget()
            datasheet.dbt_leaf.hasview = False
            self.vtgui.updateActions()
            datasheet.close()
            layout.deleteLater()

        self.deleteLater()
        QtWidgets.QMdiSubWindow.closeEvent(self, event)
        if self.vtgui.workspace.subWindowList() == []:
            self.vtgui.dbs_tree_view.setFocus(True)

    def ungroupArrays(self):
        """Break-up a GroupedArrays object into its individual components.
        """

        for view in self.vtgui.workspace.subWindowList():
            if isinstance(view, GroupedArrays):
                self = view
                break
        sw_layout = self.widget().layout()
        vl_count = sw_layout.count()
        for i in range(vl_count):
            layout = sw_layout.itemAt(i)
            title = layout.itemAt(0).widget().text()
            datasheet = layout.itemAt(1).widget()
            datasheet.setWindowFlags(QtCore.Qt.SubWindow)
            datasheet.setWindowTitle(title)
            datasheet.setParent(self.vtgui.workspace)
            datasheet.widget().verticalScrollBar().show()
            datasheet.widget().verticalHeader().show()
            cb = QtWidgets.QCheckBox(datasheet)
            cb.setToolTip(translate('GroupedArrays',
                                    "Group Arrays into a unique view",
                                    'Checkbox tooltip'))
            datasheet.leaf_view.setCornerWidget(cb)
            # Attributes for retrieving the state of the checkbox from the
            # datasheet being customized
            cb.source_array = datasheet
            datasheet.leaf_view.cornerWidget().setCheckState(QtCore.Qt.Checked)
            datasheet.is_checked = \
                datasheet.leaf_view.cornerWidget().checkState()
            datasheet.show()
            layout.deleteLater()
        self.deleteLater()
        self.vtgui.workspace.removeSubWindow(self)



"""Test class for vtgui.py"""

import pytest

from qtpy import QtWidgets
from qtpy import QtCore
from qtpy import QtGui


@pytest.mark.usefixtures('launcher')
class TestLogger(object):
    @pytest.fixture()
    def logger(self, launcher):
        return launcher.gui.logger

    def test_write(self, logger):
        text_cursor = logger.textCursor()

        # Log regular message
        logger.clear()
        logger.write('Test text')
        assert logger.toPlainText() == 'Test text'

        # Log error message
        logger.clear()
        logger.write('\nError: test error text color')
        fmt = text_cursor.charFormat()
        color = fmt.foreground().color().name()
        assert color == '#ff0000'

        # Log warning message
        logger.clear()
        logger.write('\nWarning: test warning text color')
        fmt = text_cursor.charFormat()
        color = fmt.foreground().color().name()
        assert color == '#f38908'

    def test_setupContextMenu(self, launcher):
        cm = launcher.gui.findChild(QtWidgets.QMenu, 'logger_context_menu')
        menu_actions = cm.actions()
        assert cm

        actions = [a.objectName() for a in menu_actions
                   if not (a.isSeparator() or a.menu())]
        expected_actions = ['logger_copy_action', 'logger_clear_action',
                            'logger_select_action']
        assert sorted(actions) == sorted(expected_actions)

        separators = [a for a in menu_actions if a.isSeparator()]
        assert len(separators) == 1

    def test_updateContextMenu(self, logger):
        # Copy action enabled
        logger.clear()
        logger.write('Sample text')
        logger.selectAll()
        logger.updateContextMenu()
        assert logger.copy_action.isEnabled()

        # Copy action disabled
        text_cursor = logger.textCursor()
        text_cursor.clearSelection()
        # Update the visible cursor (see qtextedit.html#textCursor docs)
        logger.setTextCursor(text_cursor)
        logger.updateContextMenu()
        assert not logger.copy_action.isEnabled()

        # Clear/select action enabled
        logger.clear()
        logger.write('Sample text')
        logger.updateContextMenu()
        assert logger.clear_action.isEnabled()
        assert logger.select_action.isEnabled()

        # Clear/select action disabled
        logger.clear()
        logger.updateContextMenu()
        assert not logger.clear_action.isEnabled()
        assert not logger.select_action.isEnabled()

    def test_focusInEvent(self, launcher, logger):
        # Give keyboard focus to the logger
        # Notice that logger.setFocus() doesn't work if the widget is hidden
        launcher.app.sendEvent(logger, QtGui.QFocusEvent(QtCore.QEvent.FocusIn))
        # QTest.mouseClick(logger, QtCore.Qt.LeftButton)

        # Check the logger configuration
        assert logger.lineWidth() == 2
        assert logger.frameShape() == QtWidgets.QFrame.Panel
        assert logger.frameShadow() == QtWidgets.QFrame.Plain

    def test_focusOutEvent(self, logger, launcher):
        # Remove keyboard focus from the logger
        launcher.app.sendEvent(logger,
                               QtGui.QFocusEvent(QtCore.QEvent.FocusOut))

        # Check the logger configuration
        assert logger.lineWidth() == logger.frame_style['lwidth']
        assert logger.frameShape() == logger.frame_style['shape']
        assert logger.frameShadow() == logger.frame_style['shadow']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module implements a buffer used to access the real data contained in
`PyTables` datasets.
By using this buffer we speed up the access to the stored data. As a
consequence, views (widgets showing a tabular representation of the dataset)
are painted much faster too.
"""

import logging
import warnings

import numpy
import tables

from qtpy import QtWidgets

from .. import utils as vtutils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# Restrict the available flavors to 'numpy' so that reading a leaf
# always return a numpy array instead of an object of the kind indicated
# by the leaf flavor. For VLArrays the read data is returned as a list whose
# elements will be numpy arrays.
tables.restrict_flavors(keep=['numpy'])
warnings.filterwarnings('ignore', category=tables.FlavorWarning)
warnings.filterwarnings('ignore', category=tables.NaturalNameWarning)

log = logging.getLogger(__name__)


class Buffer(object):
    """Buffer used to access the real data contained in `PyTables` datasets.

    Note that the buffer number of rows **must** be at least equal to
    the number of rows of the table widget it is going to fill. This
    way we avoid to have partially filled tables.  Also note that rows
    in buffer are numbered from 0 to N (as it happens with the data
    source).

    Leaves are displayed in MxN table widgets:
    - scalar arrays are displayed in a 1x1 table.
    - 1D arrays are displayed in a Mx1 table.
    - KD arrays are displayed in a MxN table.
    - tables are displayed in a MxN table with one field per column

    Implementing a reader method for each case allows us to choose
    at document creation time which method will be used. It works
    *much* faster than a global reader method that has to decide
    which block of code must be executed at every cell painting time.

    :Parameter leaf:
        the data source (`tables.Leaf` instance) from which data are
        going to be read.
    """

    def __init__(self, leaf):
        """
        Initializes the buffer.
        """

        self.leaf = leaf
        # The structure where read data will be stored.
        self.chunk = numpy.array([])

        # The method used for reading data depends on the kind of node.
        # Setting the reader method at initialization time increases the
        # speed of reading several orders of magnitude
        if isinstance(leaf, tables.Table):
            # Dataset elements will be read like a[row][column]
            self.getCell = self.arrayCell
        elif isinstance(leaf, tables.EArray):
            self.getCell = self.EArrayCell
        elif isinstance(leaf, tables.VLArray):
            # Array elements will be read like a[row]
            self.getCell = self.vectorCell
        elif leaf.shape == ():
            # Array element will be read like a[()]
            self.getCell = self.scalarCell
        elif len(leaf.shape) == 1:
            # Array elements will be read like a[row]
            self.getCell = self.vectorCell
        elif len(leaf.shape) > 1:
            # Dataset elements will be read like a[row][column]
            self.getCell = self.arrayCell

    def __del__(self):
        """Release resources before destroying the buffer.
        """
        # FIXME: PY3.5+ leaks resources (use finalizer instead).
        self.chunk = None

    def total_nrows(self):
        """Estimates the number of rows of the dataset being read.

        We don't use the `Leaf.nrows` attribute because it is not always
        suitable for displaying the data in a 2D grid. Instead we use the
        `Leaf.shape` attribute and map it to a number of rows useful for our
        purposes.

        The returned number of rows may differ from that returned by the
        `nrows` attribute in scalar arrays and `EArrays`.

        :Returns: the size of the first dimension of the dataset
        """

        shape = self.leaf.shape
        if shape is None:
            # Node is not a Leaf or there was problems getting the shape
            nrows = 0
        elif shape == ():
            # Node is a rank 0 array (e.g. numpy.array(5))
            nrows = 1
        else:
            nrows = self.leaf.nrows

        return nrows

    def readBuffer(self, start, stop):
        """
        Read a chunk from the data source.

        The size of the chunk is given in number of rows, and might
        be smaller than the requested one if the beginning/end of the
        document is reached when reading.

        Data read from `VLArrays` are returned as a Python list of objects of
        the current flavor (usually ``numpy`` arrays). Any
        other kind of `tables.Leaf` returns a ``numpy`` array (see comments on
        restricted_flavors above)

        :Parameters:
        :param start: the document row that is the first row of the chunk.
        :param stop: the last row to read, inclusive.
        """

        try:
            # data_source is a tables.Table or a tables.XArray
            # but data is a numpy array
            # Warning: in a EArray with shape (2,3,3) and extdim attribute
            # being 1, the read method will have 3 rows. However, the numpy
            # array returned by EArray.read() will have only 2 rows
            data = self.leaf.read(start, stop)
        except tables.HDF5ExtError as e:
            log.error(
                translate('Buffer', """\nError: problems reading records. """
                          """The dataset maybe corrupted.\n{}""",
                          'A dataset readability error').format(e.message))
        except:
            vtutils.formatExceptionInfo()
        else:
            # Update the buffer contents and its start position
            self.chunk = data

    def scalarCell(self, row, col):
        """
        Returns a cell of a scalar array view.

        The indices values are not checked (and could not be in the
        buffer) so they should be checked by the caller methods.

        :Parameters:
        - `row`: the row to which the cell belongs.
        - `col`: the column to wich the cell belongs

        :Returns: the cell at position `(row, col)` of the document
        """
        return self.chunk[()]

    def vectorCell(self, row, col):
        """
        Returns a cell of a 1D-array view.

        VLArrays, for instance, always are read with this method. If no
        pseudo atoms are used then a list of arrays is returned. If vlstring
        is used then a list of raw bytes is returned. If vlunicode is used
        then a list of unicode strings is returned. If object is used then a
        list of Python objects is returned.

        The indices values are not checked (and could not be in the
        buffer) so they should be checked by the caller methods.

        :Parameters:
        - `row`: the row to which the cell belongs.
        - `col`: the column to wich the cell belongs

        :Returns: the cell at position `(row, col)` of the document
        """

        # The row-coordinate must be shifted by model.start units in order to
        # get the right chunk element.
        # chunk = [row0, row1, row2, ..., rowN]
        # and columns can be read from a given row using indexing notation
        return self.chunk[row]

    def EArrayCell(self, row, col):
        """
        Returns a cell of a EArray view.

        The indices values are not checked (and could not be in the
        buffer) so they should be checked by the caller methods.

        :Parameters:
        - `row`: the row to which the cell belongs.
        - `col`: the column to wich the cell belongs

        :Returns: the cell at position `(row, col)` of the document
        """

        # The row-coordinate must be shifted by model.start units in order to
        # get the right chunk element.
        # chunk = [row0, row1, row2, ..., rowN]
        # and columns can be read from a given row using indexing notation
        # Get data for cell
        cell_data = numpy.take(self.chunk, [row], axis=self.leaf.maindim)
        # Remove extra dimension
        cell_data = numpy.squeeze(cell_data, axis=self.leaf.maindim)
        return cell_data

    def arrayCell(self, row, col):
        """
        Returns a cell of a ND-array view or a table view.

        The indices values are not checked (and could not be in the
        buffer) so they should be checked by the caller methods.

        :Parameters:
        - `row`: the row to which the cell belongs.
        - `col`: the column to wich the cell belongs

        :Returns: the cell at position `(row, col)` of the document
        """

        # The row-coordinate must be shifted by model.start units in order to
        # get the right chunk element.
        # For arrays we have
        # chunk = [row0, row1, row2, ..., rowN]
        # and columns can be read from a given row using indexing notation
        # For tables we have
        # chunk = [nestedrecord0, nestedrecord1, ..., nestedrecordN]
        # and fields can be read from nestedrecordJ using indexing notation
        return self.chunk[row][col]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Storing time series created with scikits.timeseries module in PyTables.
Example 3.
"""

import os

import numpy as np

import scikits.timeseries as ts
import scikits.timeseries.lib.tstables as tstab

# Generate random data
data = np.cumprod(1 + np.random.normal(0, 1, 300)/100)
series = ts.time_series(data, start_date=ts.Date(freq='M', year=1982, month=1))

# Write to a PyTables file
output_dir = '../timeseries'
try:
    os.mkdir(output_dir)
except OSError:
    pass

hdf5_name = 'scikits_test3.hdf5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
h5file = tstab.open_file(filepath_hdf5, mode="w",
title='Example table with csikits time series')
group_doc = h5file.create_group("/", 'examples', 'Test Data')
table = h5file.createTimeSeriesTable(group_doc, 'Example_3', series)
h5file.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Storing time series created with scikits.timeseries module in PyTables.
Example 1.
"""

import os

import numpy as np

import scikits.timeseries as ts
import scikits.timeseries.lib.tstables as tstab
import scikits.timeseries.lib.reportlib as rl

## a test series
data = np.arange(1, 366)
startdate = ts.Date(freq='D', year=2009, month=1, day=1)
test_series = ts.time_series(data, start_date=startdate, freq='D')

output_dir = '../timeseries'
try:
    os.mkdir(output_dir)
except OSError:
    pass

## write to csv file
csv_file = os.path.abspath(os.path.join(output_dir, 'test_series.csv'))
mycsv = open(csv_file, 'w')
csvReport = rl.Report(test_series, delim=';',
fixed_width=True)
csvReport(output=mycsv)
mycsv.close()

# Write to a PyTables file
hdf5_name = 'scikits_test1.hdf5'
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
h5file = tstab.open_file(filepath_hdf5, mode="w",
title='Example table with csikits time series')
group_doc = h5file.create_group("/", 'examples', 'Test Data')
table = h5file.createTimeSeriesTable(group_doc, 'Example_1', test_series)
h5file.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a dialog for renaming nodes of the tree of databases view.

The dialog is raised when a new group node is created and when an existing node
is being renamed.
"""

import os.path

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.uic import loadUiType

import vitables.utils

__docformat__ = 'restructuredtext'

# This method of the PyQt5.uic module allows for dinamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_InputNodenameDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'nodename_dlg.ui'))[0]


class InputNodeName(QtWidgets.QDialog, Ui_InputNodenameDialog):
    """
    Dialog for interactively entering a name for a given node.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This dialog is called when a new group is being created and also when
    a node of any kind is being renamed.

    Regular Qt class QInputDialog is not used because, at least apparently,
    it doesn't provide a way for customizing buttons text.

    :Parameters:

    - `title`: the dialog title
    - `info`: information to be displayed in the dialog
    - `action`: string with the editing action to be done, Create or Rename
    """

    def __init__(self, title, info, action, default_value=None):
        """A customised `QInputDialog`.
        """

        vtapp = vitables.utils.getVTApp()
        # Makes the dialog and gives it a layout
        super(InputNodeName, self).__init__(vtapp.gui)
        self.setupUi(self)

        # Set dialog caption and label with general info
        self.setWindowTitle(title)
        self.generalInfo.setText(info)

        # The Create/Rename button
        self.edit_button = self.buttonsBox.addButton(
            action, QtWidgets.QDialogButtonBox.AcceptRole)

        # Setup a validator for checking the entered node name
        validator = QtGui.QRegExpValidator(self)
        pattern = QtCore.QRegExp("[a-zA-Z_]+[0-9a-zA-Z_ ]*")
        validator.setRegExp(pattern)
        self.valueLE.setValidator(validator)

        if default_value:
            self.valueLE.setText(default_value)
            
        # Make sure that buttons are in the proper activation state
        self.valueLE.textChanged.emit(self.valueLE.text())

    @QtCore.Slot("QString", name="on_valueLE_textChanged")
    def checkName(self, current):
        """
        Check the current name value.

        The state of the `Create` button depends on the passed string. If
        it is empty then the button is disabled. Otherwise it is enabled.

        :Parameter current: the value currently displayed in the text box
        """

        if current == '':
            self.edit_button.setEnabled(0)
        else:
            self.edit_button.setEnabled(1)

    @QtCore.Slot(name="on_buttonsBox_accepted")
    def saveName(self):
        """Slot for saving the entered name and closing the dialog.
        """

        self.node_name = self.valueLE.text()
        self.accept()
